/*
    orgasm - A machine code assembler for 6502 microprocessors 
    Copyright (C) 1999 Robert Norris

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Report bugs to: rob@nauseum.org
*/

/*
 * eval.c - the expression parser
 */


#include "config.h"

#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#ifdef HAVE_DMALLOC_H
 #include <dmalloc.h>
#endif

#include "types.h"
#include "proto.h"

/* stack for conversions */
int *stack = NULL;
int s_end = 0;

/* temp area for the postfix expression */
int *postfix = NULL;
int p_end = 0;

/* free the stack */
inline void s_free(void) {
	if(stack)
		free(stack);
	stack = NULL;
	s_end = 0;
	}

/* push a value onto the stack */
void s_push(int n) {
	if(!(stack = (int *) realloc(stack, sizeof(int) * (s_end + 1))))
		do_oom();
	stack[s_end] = n;
	s_end++;
	}

/* pop a value off the stack */
int s_pop(void) {
	int n;
	if(s_end) {
		s_end--;
		n = stack[s_end];
		if(s_end) {
			if(!(stack = (int *) realloc(stack, sizeof(int) * s_end)))
				do_oom();
			}
		else
			s_free();
		return n;
		}
	else
		return 0;
	}

/* get the value on top of the stack */
inline int s_top(void) {
	if(s_end)
		return stack[s_end - 1];
	else
		return 0;
	}

/* clear the postfix expr */
inline void p_free(void) {
	if(postfix)
		free(postfix);
	postfix = NULL;
	p_end = 0;
	}

/* add a value onto the end of the postfix expr */
void p_add(int n) {
	if(!(postfix = realloc(postfix, sizeof(int) * (p_end + 1))))
		do_oom();
	postfix[p_end] = n;
	p_end++;
	}

/* see if this char is a valid operator */
int isop(int c) {
	const char ops[] = "+-*/()[]";
	int i;
	
	for(i = 0; i < 8; i++)
		if(ops[i] == c)
			return 1;
	
	return 0;
	}

/* get the next 'token' from an infix expression */
char *get_expr_tok(char **expr, u16 fakeubound) {
	char *new = NULL, *temp;
	struct symbol *sym;
	int i;

	/* empty expr */
	if(!**expr)
		return NULL;
	
	/* get rid of the spaces */
	if(isspace(**expr)) {
		(*expr)++;
		new = NULL;
		}
								
	else if(((*expr)[0] == '*') && ((*expr)[1] == '\0')) {
		if(!(new = (char *) malloc(sizeof(char) * 6)))
			do_oom();
		sprintf(new, "$%x", loc);
		(*expr)++;
		}
		
	/* next token is an operator */
	else if(isop(**expr)) {
		if(!(new = (char *) malloc(sizeof(char) * 2)))
			do_oom();
		new[0] = **expr;
		new[1] = '\0';
		(*expr)++;
		}
		
	/* next token is a number */
	else if((**expr == '$') || (**expr == '%') || isdigit(**expr)) {

		/* work out the type, and grab all its digits */
		if(**expr == '$')
			for(i = 1; isxdigit((*expr)[i]); i++);
		else if(**expr == '%')
			for(i = 1; ((*expr)[i] == '0') || ((*expr)[i] == '1'); i++);
		else
			for(i = 1; isdigit((*expr)[i]); i++);
		if(!isop((*expr)[i]) && !isspace((*expr)[i]) && ((*expr)[i] != '\0'))
			goto err_op;
		if(!(new = (char *) malloc(sizeof(char) * (i + 1))))
			do_oom();
		strncpy(new, *expr, i);
		new[i] = '\0';
		*expr = &(*expr)[i];
		}

	/* next token is a symbol */
	else if(issymchar(**expr) && !isdigit(**expr)) {

		/* grab all the symbol chars */
		for(i = 1; issymchar((*expr)[i]); i++);
		if(!isop((*expr)[i]) && !isspace((*expr)[i]) && ((*expr)[i] != '\0'))
			goto err_op;
		if(!(temp = strdup(*expr)))
			do_oom();
		temp[i] = '\0';
		if(!(new = (char *) malloc(sizeof(char) * 6)))
			do_oom();

		/* this really sucks, building a string which will
		   promptly be turned back into an int when we return */
		if((sym = get_symbol(temp)) && sym->defined)
			sprintf(new, "$%x", sym->value);
		else if(fakeubound > 0xff) {
			strcpy(new, "$8000\0");
			add_symbol(temp);
			}
		else {
			strcpy(new, "$80\0");
			add_symbol(temp);
			}
		*expr = *expr + strlen(temp);
		free(temp);
		}
	
	/* bogus char */
	else
		goto err_ill;

	return new;

err_op:
	do_error("operator expected\n");
err_ill:
	do_error("illegal character in expression\n");

	return NULL;
	}

/* atoi with three different types .. do we need octal as well? */
long int new_atoi(char *expr, long int ubound) {
	long int val = 0;
	char *end;

	/* convert it based on type */
	if(*expr == '$')
		val = strtol(&expr[1], &end, 16);
	else if(*expr == '%')
		val = strtol(&expr[1], &end, 2);
	else if(isdigit(*expr))
		val = strtol(expr, &end, 10);
	else
		do_error("number format is unrecognized\n");

	if(*end)
		do_error("number contains illegal digit\n");

	if(ubound && (val > ubound))
		do_error("number out of bounds\n");

	return val;
	}

/* infix -> postfix */
void eval_infix(char *expr, u16 fakeubound) {
	char *token = NULL;
	int f_op = 0, op;

	do {
		if(token) free(token);
		token = get_expr_tok(&expr, fakeubound);
		} while(!token && *expr);
					
	while(token) {
		switch(*token) {
			case '+': case '-':
				if((s_top() != ('(' | 0x100)) && f_op)
					p_add(s_pop());
				else
					f_op = 1;
				s_push(*token | 0x100);
				break;

			case '*': case '/':
				if((s_top() != ('(' | 0x100)) && f_op && ((s_top() == ('*' | 0x100)) || (s_top() == ('/' | 0x100))))
					p_add(s_pop());
				else
					f_op = 1;
				s_push(*token | 0x100);
				break;

			case '(': case '[':
				if(!f_op)
					f_op = 1;
				s_push('(' | 0x100);
				break;

			case ')': case ']':
				op = s_pop();
				while(op != ('(' | 0x100)) {
					p_add(op);
					op = s_pop();
					}
				break;

			default:
				p_add(new_atoi(token, 0));
				break;
			}

		do {
			if(token) free(token);
			token = get_expr_tok(&expr, fakeubound);
			} while(!token && *expr);
		}		

	while(s_end)
		p_add(s_pop());
	
	s_free();
	}

int eval_postfix(void) {
	int i, a, b;

	for(i = 0; i < p_end; i++) {
		switch(postfix[i]) {
			case ('+' | 0x100):
				b = s_pop(); a = s_pop();
				s_push(a + b);
				break;
			case ('-' | 0x100):
				b = s_pop(); a = s_pop();
				s_push(a - b);
				break;
			case ('*' | 0x100):
				b = s_pop(); a = s_pop();
				s_push(a * b);
				break;
			case ('/' | 0x100):
				b = s_pop(); a = s_pop();
				if(!b)
					do_error("division by zero\n");
				s_push(a / b);
				break;
			default:
				s_push(postfix[i]);
				break;
			}
		}
	
	a = s_pop();

	if(s_end)
		do_error("operator expected\n");

	s_free();
	p_free();

	return a;
	}

u16 eval(char *expr, u16 ubound) {
	int type = 0, val;

	chop(&expr);

	/* low byte */
	if(*expr == '<') {
		type = 1;
		(expr)++;
		}
	
	/* high byte */
	else if(*expr == '>') {
		type = 2;
		(expr)++;
		}
	
	eval_infix(expr, ubound);
	val = eval_postfix();

	if(!type && (val > ubound)) 
		do_error("result of expression is out of bounds\n");

	/* grab the correct bit*/
	if(type == 1)
		val &= 0xff;
	else if(type == 2)
		val >>= 8;
	
	return val;
	}
