/*
    orgasm - A machine code assembler for 6502 microprocessors 
    Copyright (C) 1999 Robert Norris

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Report bugs to: rob@nauseum.org
*/

#include "config.h"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include "types.h"
#include "proto.h"

struct opcode ops[] = {
	{"ADC", 0, 0, 0x69, 0x65, 0x75, 0, 0x6d, 0x7d, 0x79, 0, 0x61, 0x71},
	{"AND", 0, 0, 0x29, 0x25, 0x35, 0, 0x2d, 0x3d, 0x39, 0, 0x21, 0x31},
	{"ASL", 0x0a, 0, 0, 0x06, 0x16, 0, 0x0e, 0x1e, 0, 0, 0, 0},
	{"BCC", 0, 0x90, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"BCS", 0, 0xB0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"BEQ", 0, 0xF0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"BIT", 0, 0, 0, 0x24, 0, 0, 0x2c, 0, 0, 0, 0, 0},
	{"BMI", 0, 0x30, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"BNE", 0, 0xD0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"BPL", 0, 0x10, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"BRK", 0x00, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"BVC", 0, 0x50, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"BVS", 0, 0x70, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"CLC", 0x18, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"CLD", 0xd8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"CLI", 0x58, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"CLV", 0xb8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"CMP", 0, 0, 0xc9, 0xc5, 0xd5, 0, 0xcd, 0xdd, 0xd9, 0, 0xc1, 0xd1},
	{"CPX", 0, 0, 0xe0, 0xe4, 0, 0, 0xec, 0, 0, 0, 0, 0},
	{"CPY", 0, 0, 0xc0, 0xc4, 0, 0, 0xcc, 0, 0, 0, 0, 0},
	{"DEC", 0, 0, 0, 0xc6, 0xd6, 0, 0xce, 0xde, 0, 0, 0, 0},
	{"DEX", 0xca, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"DEY", 0x88, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"EOR", 0, 0, 0x49, 0x45, 0x55, 0, 0x4d, 0x5d, 0x59, 0, 0x41, 0x51},
	{"INC", 0, 0, 0, 0xe6, 0xf6, 0, 0xee, 0xfe, 0, 0, 0, 0},
	{"INX", 0xe8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"INY", 0xc8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"JMP", 0, 0, 0, 0, 0, 0, 0x4c, 0, 0, 0x6c, 0, 0},
	{"JSR", 0, 0, 0, 0, 0, 0, 0x20, 0, 0, 0, 0, 0},
	{"LDA", 0, 0, 0xa9, 0xa5, 0xb5, 0, 0xad, 0xbd, 0xb9, 0, 0xa1, 0xb1},
	{"LDX", 0, 0, 0xa2, 0xa6, 0, 0xb6, 0xae, 0, 0xbe, 0, 0, 0},
	{"LDY", 0, 0, 0xa0, 0xa4, 0xb4, 0, 0xac, 0, 0xbc, 0, 0, 0},
	{"LSR", 0x4a, 0, 0, 0x46, 0x56, 0, 0x4e, 0x5e, 0, 0, 0, 0},
	{"NOP", 0xea, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"ORA", 0, 0, 0x09, 0x05, 0x15, 0, 0x0d, 0x1d, 0x19, 0, 0x01, 0x11},
	{"PHA", 0x48, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"PHP", 0x08, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"PLA", 0x68, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"PLP", 0x28, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"ROL", 0x2a, 0, 0, 0x26, 0x36, 0, 0x2e, 0x3e, 0, 0, 0, 0},
	{"ROR", 0x6a, 0, 0, 0x66, 0x76, 0, 0x6e, 0x7e, 0, 0, 0, 0},
	{"RTI", 0x40, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"RTS", 0x60, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"SBC", 0, 0, 0xe9, 0xe5, 0xf5, 0, 0xed, 0xfd, 0xf9, 0, 0xe1, 0xf1},
	{"SEC", 0x38, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"SED", 0xf8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"SEI", 0x78, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"STA", 0, 0, 0, 0x85, 0x95, 0, 0x8d, 0x9d, 0x99, 0, 0x81, 0x91},
	{"STX", 0, 0, 0, 0x86, 0, 0x96, 0x8e, 0, 0, 0, 0, 0},
	{"STY", 0, 0, 0, 0x84, 0x94, 0, 0x8c, 0, 0, 0, 0, 0},
	{"TAX", 0xaa, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"TAY", 0xa8, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"TSX", 0xba, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"TXA", 0x8a, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"TXS", 0x9a, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{"TYA", 0x98, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0},
	{0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}
	};

struct pseudo pseudos[] = {
	{".org", P_ORG},
	{".byte", P_BYTE},
	{".word", P_WORD},
	{".ascii", P_ASCII},
	{".export", P_EXPORT},
	{0, 0}
	};

struct opcode *get_opcode(char *name) {
	struct opcode *scan;
	
	for(scan = ops; scan->name; scan++)
		if(!strcasecmp(scan->name, name))
			return scan;
	
	return NULL;
	}

struct pseudo *get_pseudo(char *name) {
	struct pseudo *scan;

	for(scan = pseudos; scan->name; scan++)
		if(!strcasecmp(scan->name, name))
			return scan;

	return NULL;
	}

struct instruction *get_ins(char *opcode, char *operand) {
	static struct instruction ins;
	struct opcode *op;
	char *scan;
	u16 val;

	/* get the opcode */
	op = get_opcode(opcode);

	/* no operand = implied mode */
	if(!operand) {
		if(!(ins.opcode = op->implied) && strcmp(op->name, "BRK"))
			do_error("'%s' has no implied mode\n", op->name);
		ins.length = 1;
		return &ins;
		}

	scan = operand;

	/* some sort of indirect addressing mode */
	if(scan[0] == '(') {
		if(endswith(&scan[1], ",x)")) {
			if(!(ins.opcode = op->indirectx))
				goto err_mode;
			
			cut(&scan[1], 3);
			val = eval(&scan[1], 0xff);
			
			ins.oper_low = val;
			ins.length = 2;
			
			return &ins;
			}
		if(endswith(&scan[1], "),y")) {
			if(!(ins.opcode = op->indirecty))
				goto err_mode;
			
			cut(&scan[1], 3);
			val = eval(&scan[1], 0xff);
			
			ins.oper_low = val;
			ins.length = 2;
			
			return &ins;
			}
		if(endswith(&scan[1], ")")) {
			if(!(ins.opcode = op->indirect))
				goto err_mode;
			
			cut(&scan[1], 1);
			val = eval(&scan[1], 0xffff);
			
			ins.oper_low = val & 0xff;
			ins.oper_high = val >> 8;
			ins.length = 3;
			
			return &ins;
			}
		}
	
	/* a hash means immediate mode */
	if(scan[0] == '#') {
		if(!(ins.opcode = op->immediate))
			goto err_mode;
			
		val = eval(&scan[1], 0xff);
			
		ins.oper_low = val;
		ins.length = 2;
			
		return &ins;
		}
	
	/* might be a branch */
	if(op->relative) {
		val = eval(scan, 0xffff);

		ins.oper_low = (val - (loc + 2)) & 0xff;

		ins.opcode = op->relative;
		ins.length = 2;

		return &ins;
		}
	
	/* absolute,x or zeropage,x */
	if(endswith(scan, ",x")) {
		cut(scan, 2);
		val = eval(scan, 0xffff);

		if((val <= 0xff) && ((ins.opcode = op->zeropagex))) {
			ins.oper_low = val;
			ins.length = 2;
			return &ins;
			}
		if((ins.opcode = op->absolutex)) {
			ins.oper_low = val & 0xff;
			ins.oper_high = val >> 8;
			ins.length = 3;
			return &ins;
			}
			
		goto err_mode;
		}

	/* absolute,y or zeropage,y */
	if(endswith(scan, ",y")) {
		cut(scan, 2);
		val = eval(scan, 0xffff);

		if((val <= 0xff) && ((ins.opcode = op->zeropagey))) {
			ins.oper_low = val;
			ins.length = 2;
			return &ins;
			}
		if((ins.opcode = op->absolutey)) {
			ins.oper_low = val & 0xff;
			ins.oper_high = val >> 8;
			ins.length = 3;
			return &ins;
			}
			
		goto err_mode;
		}
	
	/* anything else is absolute or zeropage */
	val = eval(scan, 0xffff);

	if((val <= 0xff) && ((ins.opcode = op->zeropage))) {
		ins.oper_low = val;
		ins.length = 2;
		return &ins;
		}
	if((ins.opcode = op->absolute)) {
		ins.oper_low = val & 0xff;
		ins.oper_high = val >> 8;
		ins.length = 3;
		return &ins;
		}

err_mode:
	do_error("this instruction does not support this addressing mode\n");
	
	return NULL;
	}
