/*
 * $Header: /home/paul/cvsroot/ver/behave.h,v 1.24 1999/12/18 03:13:07 paul Exp $
 *
 * Copyright (c) Paul Stravers 1999.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */


/*
 * The enumerated type "stmt_type" lists all possible statements that we
 * recognize in procedural blocks like Verilog "always" and "initial".
 */
enum stmt_type {
   Stmt_Null   = 0,             /* empty statement */
   Stmt_Assign = 1,             /* procedural blocking assignment */
   Stmt_Group  = 2,             /* begin ... end grouping statement */
   Stmt_Branch = 3,             /* if-else statement */
   Stmt_Case   = 4              /* case ... endcase statement */
};

typedef unsigned long stmt_flags; /* flags that modify the stmt semantics */

struct regval
{
   struct symslot *symbol;     /* the lexical name of this register */
   struct muxtree *muxtree;     /* the current value of this register */
   struct regval  *next;
};

struct reg_lval
{
   struct reg       *reg;
   struct bit_range *bit_range;
   struct reg_lval  *next;
};

/*
 * the following enum is used to test values in the stmt.flags field:
 */
enum {
   Null_Flag = 0,
   Case_Parallel = 0x1,         /* Do not prioritize choices in Stmt_Case
                                   (as in //pragma parallel_case) */
   Case_With_Dontcare = 0x2,    /* accept X, Z and ? values in choice list */
   Mesg_Dontcare = 0x4,         /* whether err mesg has been printed  */
   Is_Async_Latch = 0x8,        /* Stmt_Branch used for async set/reset */
   Is_Non_Blocking = 0x10       /* Stmt_Assign is of non-blocking type */
};

/*
 * The structure "stmt" contains all relevant information about a statement
 * that can occur inside a Verilog "always" block.
 */
struct stmt
{
   enum stmt_type type;         /* which type of statement this is */
   stmt_flags flags;            /* flags to detail the statement semantics */
   struct regval *lvalues;      /* list of regs that this statement assigns
                                 * to.  This is the union of all assignments
                                 * of this statements and all its
                                 * sub-statements, e.g. if this is a
                                 * Stmt_Branch then it is the union of the
                                 * reg assignments in both branches of the
                                 * if-else statement, recursively.
                                 */
   union
   {                            /* use stmt_type to select valid union member*/
      struct                    /* stmt_type is Stmt_Assign */
      {
         struct assign_reg *reg_list;
      } assign;
      struct                    /* stmt_type is Stmt_Group */
      {
         struct stmt *first;    /* list of statements grouped by Stmt_Group, */
         struct stmt *last;     /* with pointers to the first and last stmt */
      } group;
      struct                    /* stmt_type is Stmt_Branch */
      {
         struct bitvec *cond;   /* the condition for selecting a branch */
         struct stmt *if_true;  /* true and false branches of Stmt_Branch */
         struct stmt *if_false; /* the "else" part can be empty */
      } branch;
      struct                    /* stmt_type is Stmt_Case */
      {
         struct bitvec *expr;   /* the expression to match against choices */
         struct bitvec *guard;  /* the booleans that guard the choices */
         struct choice *list;   /* list of choices inside Stmt_Case */
         struct choice *deflt;  /* the default choice (may be empty) */
      } choice;
   } t;

   int lineno;                  /* line number where this statement appears */
   struct stmt *parent;         /* Stmt_Group/If/Case that encloses us */
   struct stmt *prev, *next;    /* next/prev statement in parent Group */
};

struct assign_reg
{
   struct symslot *reg;   /* the register we are assigning to */
   struct bitvec *expr;   /* the expression being assigned */
   int hi_range_offset;   /* when assigning to a sub-range, this
                           * is the diff between hirange and the
                           * highest index of the subrange */
   struct assign_reg *next;
};

/*
 * The structure "choice" holds the {selector,statement} pair of a
 * choice in a case statement.
 */
struct choice
{
   struct choice_selector *sel; /* The constant(s) that selects this choice */
   struct stmt* statement;      /* The statement executed when selected */
   int lineno;                  /* line number where this choice appears */
   char msg_size;               /* only print error message once */
   char msg_const;
   struct choice* next;         /* link to next choice in Stmt_Case */
};

/*
 * Every arm in the Case switch can be selected by one or more constants:
 */
struct choice_selector
{
   struct bitvec* const_expr;
   struct choice_selector *next;
};

/*
 * This holds a range
 */
struct bit_range
{
   int left, right;
};

/*
 * A list of symbols:
 */
enum SymListBits { SYML_POSEDGE_B,
                   SYML_NEGEDGE_B,
                   SYML_ASYNC_B};	/* these are for symlist.flags */
#define SYML_POSEDGE (1<<SYML_POSEDGE_B)
#define SYML_NEGEDGE (1<<SYML_NEGEDGE_B)
#define SYML_ASYNC   (1<<SYML_ASYNC_B)

struct symlist
{
   struct symslot *symbol;
   int flags;
   struct symlist *next;
};


struct stmt *new_stmt(enum stmt_type type, int lineno);
struct stmt *stmt_assign_expr(struct stmt *this,
                              int llorange, int lhirange,
                              struct reg *lval,
                              struct bitvec *rval);
struct stmt *stmt_group_append(struct stmt *this,
                               struct stmt *newstmt);
struct stmt *stmt_branch_true(struct stmt *this,
                              struct bitvec *condition,
                              struct stmt *truestmt);
struct stmt *stmt_branch_false(struct stmt *this,
                               struct stmt *elsestmt);
struct stmt *stmt_case_init(struct stmt *this,
                            struct bitvec *switch_expr,
                            struct choice *choice_list);
void elaborate_always(struct stmt *this,
                      struct symlist *ev_list,
                      int evt_lineno);
struct bitvec *compare_vectors(struct bitvec *b1,
                               struct bitvec *b2,
                               int invert_output);
void verify_event_list(struct symlist *event_list,
                       struct symlist *read_list,
                       int lineno);
int pragma_expected(int new_state);
struct stmt *stmt_case_setflag(struct stmt *this,
                               stmt_flags flag);
struct choice *new_choice(struct choice_selector *sel,
                          struct stmt *statement,
                          int lineno);
struct choice_selector *new_choice_selector(struct bitvec *const_expr);
struct bitvec *new_bitvec(int length);
struct reg_lval *new_reg_lval(struct reg *reg, struct bit_range *bit_range);
void align_bitvec(struct bitvec **aptr, struct bitvec **bptr);
void split_bitvec(struct bitvec **aptr, struct bitvec **bptr, int b_size);
void resize_bitvec(struct bitvec **aptr, int new_size);
struct symslot *generate_gate(int type,
                              struct bitvec *gate_inputs,
                              int num_inputs);
struct symslot *generate_gate0(int type);
struct symslot *generate_gate1(int type,
                               struct symslot *gate_input1);
struct symslot *generate_gate2(int type,
                               struct symslot *gate_input1,
                               struct symslot *gate_input2);
struct symslot *generate_gate3(int type,
                               struct symslot *gate_input1,
                               struct symslot *gate_input2,
                               struct symslot *gate_input3);
struct bitvec *generate_adder(struct bitvec *lhs, struct bitvec *rhs,
                              int operator,
                              int sum2, int carry2, int sum3, int carry3);
struct bitvec *number_to_bitvec(int number, int length);
int get_driver_type(struct symslot* sig);

int stmt_case_testflag(struct stmt *this,
                       stmt_flags flag);

