/*
 * $Header: /home/paul/cvsroot/ver/preproc.c,v 1.6 1999/12/16 01:37:22 paul Exp $
 *
 * Copyright (c) Paul Stravers 1999.
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#include <stdio.h>
#include <stdlib.h>
#include "symbol.h"

extern struct filename *filename; /* list of verilog files to process */
extern int yylineno;
extern FILE *yyin;
extern int wwerror_count;

#define LineSz       2048
#define ExpLineSz    5*LineSz
#define MaxNestDepth 50         /* for `ifdef nesting */

#define COND_ON      0
#define COND_OFF     1
#define COND_BLOCKED 2

static struct filename *curr_fn = 0;
static char linbuf[LineSz + 2];
static char expbuf[ExpLineSz + 2];
static int pplineno;
static int cond_stack[MaxNestDepth+1];
static int cond_ptr;
static int cond;

static int preproc_lines(char *buf, int max_size);
static int preproc_new_line();
static int enter_macro(int *ii, int j, int *newlines);
static int ifdef_macro(int *ii, int *new_cond);
static int expand_macro(int *jj, int start_of_token, int end_of_token);
static int find_backquote(int *bq, int start, int eob);
static struct sym_entry *find_symbol(int *eos, int start, int eob);
static int replace_symbol(int bq, int eos, int *eob, struct sym_entry *sym);
static int is_alpha_num(int c);
static int is_space(int c);
static void strip_comment_and_trailing_space(int bov, int *eovptr);

/*
 * Verilog preprocessor. Loads up to MAX_SIZE characters in BUF.  Returns
 * number of characters actually put in BUF, or 0 if end-of-file.
 */
int preproc(char *buf, int max_size)
{
   int num_read;
   static first_call = TRUE;
   if (first_call)
   {
      first_call = FALSE;
      curr_fn = filename;
      if (curr_fn == 0)
	 {
         yyin = stdin;          /* read from stdin */
	 vpp_main(yyin, "<stdin>");
         fclose(yyin);             /* close current verilog file */
         yyin = NULL;
	 }
      else
      {
         yylineno = 1;
         yyin = fopen(curr_fn->name, "r");
         if (yyin == 0)
         {
            mesg("** Cannot open file \"%s\"\n", curr_fn->name);
            exit(1);
         }
	 vpp_main(yyin, curr_fn->name);
         warn("Processing file \"%s\" ...\n", curr_fn->name);
         fclose(yyin);             /* close current verilog file */
         yyin = NULL;
	 vpp_update_yyfilename(curr_fn->name);
      }
      cond_ptr = 0;
      cond = cond_stack[cond_ptr] = COND_ON;
   }
   num_read = preproc_lines(buf, (size_t)max_size);
   if ((num_read == 0) && (curr_fn != 0) && (wwerror_count==0) )
   {
      if (cond_ptr > 0)
         mesg("** unterminated `ifdef detected at end-of-file\n");
      cond_ptr = 0;
      cond = cond_stack[cond_ptr] = COND_ON;
      curr_fn = curr_fn->next;
      if (curr_fn != 0)
      {
         yylineno = 1;
         yyin = fopen(curr_fn->name, "r"); /* open next verilog file */
         if (yyin == 0)
         {
            mesg("** Cannot open file \"%s\"\n", curr_fn->name);
            exit(1);
         }
	 vpp_main(yyin, curr_fn->name);
         warn("Processing file \"%s\" ...\n", curr_fn->name);
         fclose(yyin);             /* close current verilog file */
         yyin = NULL;
         vpp_update_yyfilename(curr_fn->name);
         num_read = preproc_lines(buf, max_size);
      }
   }
   return num_read;
}


static int preproc_lines(char *buf, int max_size)
{
   static int exp_sz = 0;
   static int exp_current = 0;
   int chunk_sz;
   if (exp_sz == 0)
   {
      exp_sz = preproc_new_line(); /* fetch new lines */
      exp_current = 0;
   }
   if (exp_sz == 0) return 0;
   chunk_sz = max_size;
   if (chunk_sz > exp_sz)
      chunk_sz = exp_sz;
   memcpy(buf, &expbuf[exp_current], chunk_sz);
   exp_sz -= chunk_sz;
   exp_current += chunk_sz;
   return chunk_sz;
}


static int is_alpha_num(int c)
{
   return isalnum(c) || (c == '_');
}

static int is_space(int c)
{
   return isspace(c) && (c != '\n');
}


static int preproc_new_line()
{
   static int remainder_start = 0, remainder_end = 0;
   static int inside_comment = 0;
   int num_read, newlines;
   int i = 0, j = 0, k = 0, c, toklen;
   pplineno = 0;
   for (k = remainder_start; k < remainder_end; k += 1)
      linbuf[i++] = linbuf[k];  /* copy left-over from previous block */
   remainder_start = remainder_end = 0;
   /* num_read = (int)fread(&linbuf[i], sizeof(char), LineSz - i, yyin); */
   num_read = mgetbuf(&linbuf[i], LineSz - i);
   num_read += i;               /* account for left-over */
   linbuf[num_read] = 0;        /* mark end of buffer */
   /* Now scan the linbuf and substitute macros: */
   i = 0;
   while (i < num_read)
   {
      int start_of_token, end_of_token, back_quote;
      c = linbuf[i];
      if      ((inside_comment == 0) && (c == '/'))  inside_comment = 1;
      else if ((inside_comment == 1) && (c == '*'))  inside_comment = 2;
      else if ((inside_comment == 1) && (c == '/'))  inside_comment = 3;
      else if ( inside_comment == 1)                 inside_comment = 0;
      else if ((inside_comment == 2) && (c == '*'))  inside_comment = 4;
      else if ((inside_comment == 3) && (c == '\n')) inside_comment = 0;
      else if ((inside_comment == 4) && (c == '/'))  inside_comment = 0;
      else if ((inside_comment == 4) && (c == '*'))  inside_comment = 4;
      else if ( inside_comment == 4)                 inside_comment = 2;
      if (c != '`')
      {
         if (c == '\n')
         {
            pplineno += 1;
            expbuf[j++] = c;
         }
         else if (cond == COND_ON)
            expbuf[j++] = c;
         else
            expbuf[j++] = ' ';
         i += 1;
         continue;
      }
      if (inside_comment >= 2)
      {
         expbuf[j++] = c;
         i += 1;
         continue;
      }
      /* at this point we have found a back quote. Go interpret it: */
      back_quote = i;
      newlines = 0;
      do                        /* skip leading spaces */
         if (i >= num_read)
         {  /* we're falling off the end of the buffer ... */
            remainder_start = back_quote;
            remainder_end = back_quote + 1;
            return j;
         }
         else
            i += 1;
      while (((c = linbuf[i]) == ' ') || (c == '\t')
             || ((c == '\\') && (linbuf[i+1] == '\n') && ++newlines && ++i));
      pplineno += newlines;
      /* at this point we have skipped over the space after the back quote */
      start_of_token = i;
      do
         if (i >= num_read)
         {  /* we reached the end of linbuf before the token was parsed */
            remainder_start = back_quote;
            remainder_end = num_read;
            return j;
         }
         else
            i += 1;
      while (is_alpha_num(linbuf[i]));
      if (linbuf[i] == 0)
      {  /* end of linbuf, continue this macro next time */
         remainder_start = back_quote;
         remainder_end = num_read;
         return j;
      }
      end_of_token = i;
      toklen = end_of_token - start_of_token;
      if ((toklen == 0) || !is_alpha_num(linbuf[start_of_token]))
         mesg("** Illegal macro name in DEFINE, '%s' line %d\n", yyfilename, yylineno+pplineno);
      else if ((toklen == 6) &&
               (strncmp(&linbuf[start_of_token], "define", toklen) == 0))
      {  /* a definition is being entered */
         if (!enter_macro(&i, j, &newlines))
         {  /* define extends beyond end of buffer */
            remainder_start = back_quote;
            remainder_end = num_read;
            return j;
         }
      }
      else if ((toklen == 5) &&
               (strncmp(&linbuf[start_of_token], "ifdef", toklen) == 0))
      {
         int new_cond;
         if (cond_ptr >= MaxNestDepth)
         {
            mesg("** Max nesting depth for `ifdef reached (%d) in '%s' line %d\n",
                 cond_ptr, yyfilename, yylineno+pplineno);
            exit(1);
         }
         else if (cond == COND_ON)
            if (!ifdef_macro(&i, &new_cond))
            {  /* linbuf ended too early. Retry next time */
               remainder_start = back_quote;
               remainder_end = num_read;
               return j;
            }
            else
               cond = cond_stack[++cond_ptr] = new_cond;
         else
            cond = cond_stack[++cond_ptr] = COND_BLOCKED;
      }
      else if ((toklen == 4) &&
               (strncmp(&linbuf[start_of_token], "else", toklen) == 0))
      {
         if (cond_ptr <= 0)
            mesg("** `else without `ifdef in '%s' on line %d\n",yyfilename, yylineno+pplineno);
         else if (cond != COND_BLOCKED)
            if (cond == COND_ON)
               cond = cond_stack[cond_ptr] = COND_OFF;
            else
               cond = cond_stack[cond_ptr] = COND_ON;
      }
      else if ((toklen == 5) &&
               (strncmp(&linbuf[start_of_token], "endif", toklen) == 0))
      {
         if (cond_ptr <= 0)
            mesg("** unmatched `endif in '%s' on line %d\n",yyfilename, yylineno+pplineno);
         else
            cond = cond_stack[--cond_ptr];
      }
      else
      {  /* a macro is being called. Expand it. */
         if (!expand_macro(&j, start_of_token, end_of_token))
         {  /* No room left in expbuf. Retry next time */
            remainder_start = back_quote;
            remainder_end = num_read;
            return j;
         }
      }
      /* we need to keep the number of newlines in linbuf exactly equal
       * to the number of newlines in expbuf. If not, the error messages
       * that may be generated by the parser will refer to lines that do
       * not correspond to the line numbers in the user's input file.
       */
      for (k = 0; k < newlines; k += 1)
         expbuf[j++] = '\n';
   }
   return j;
}


struct sym_entry
{
   int symbol_length, value_length;
   char *symbol;
   char *value;
   struct sym_entry *next;
};

static struct sym_entry *preproc_symtab = 0;

/*
 * Enter a macro in the symbol table. The macro name is the
 * first word thet II points to, the macro value follows and
 * extends to the end of line.
 */
static int enter_macro(int *ii, int j, int *newlines)
{
   int i = *ii, c, k, bos, eos, bov, eov, nl1 = 0, nl2 = 0;
   char *sym, *val;
   struct sym_entry *sym_entry;
   while ((c = linbuf[i]) &&
          (is_space(c) ||
           ((c == '\\') && (linbuf[i+1] == '\n') && ++nl1 && ++i)))
      i += 1;                   /* eat spaces */
   pplineno += nl1;
   if (c == 0) return FALSE;    /* end of buffer */
   if (!is_alpha_num(c))
   {
      mesg("** Illegal macro name in `define: '%s' line %d\n", yyfilename, yylineno+pplineno);
      *ii = i;
      return TRUE;
   }
   bos = i;
   while ((c = linbuf[i]) && is_alpha_num(c))
      i += 1;
   if (c == 0) return FALSE;    /* end of buffer */
   eos = i;
   while ((c = linbuf[i]) &&
          (is_space(c) ||
           ((c == '\\') && (linbuf[i+1] == '\n') && ++nl2 && ++i)))
      i += 1;                   /* eat spaces */
   pplineno += nl2;
   if (c == 0) return FALSE;    /* end of buffer */
   if ((*newlines + nl1 + nl2 + j) > ExpLineSz)
      return FALSE; /* no room left for newlines in expbuf */
   bov = i;
   while ((c = linbuf[i]) && (c != '\n'))
      if ((c == '\\') && linbuf[i+1] == '\n')
         i += 2;
      else
         i += 1;
   if (c == 0) return FALSE;    /* end of buffer */
   eov = i;
   if (cond == COND_ON)
   {
      /* the symbol value starts at bov and ends at eov. We need to strip
       * comments from the value, especially the //...<newline> type of
       * comments, because when we later expand the macro, the <newline>
       * won't be present anymore and therefore the comment will not be
       * properly terminated...!!!
       */
      strip_comment_and_trailing_space(bov, &eov);
      sym = calloc(1, eos - bos + 1);
      strncpy(sym, &linbuf[bos], eos - bos);
      sym[eos - bos] = 0;
      val = calloc(1, eov - bov + 1);
      for (k = 0; k < eov - bov; k += 1)
         if (((c = linbuf[bov + k]) == '\\') && (linbuf[bov + k + 1] == '\n'))
            val[k++] = ' ', val[k] = ' '; /* erase escaped newline */
         else
            val[k] = linbuf[bov + k];
      val[k] = 0;
      sym_entry = calloc(1, sizeof(struct sym_entry));
      sym_entry->symbol = sym;
      sym_entry->value = val;
      sym_entry->symbol_length = eos - bos;
      sym_entry->value_length = eov - bov;
      if (preproc_symtab == 0)
         preproc_symtab = sym_entry;
      else
      {
         sym_entry->next = preproc_symtab;
         preproc_symtab = sym_entry;
      }
   }
   *ii = i;
   *newlines += nl1 + nl2;
   return TRUE;
}


static void strip_comment_and_trailing_space(int bov, int *eovptr)
{
   int k, eov = *eovptr, c;
   int inside_comment = 0;
   for (k = bov; k < eov; k += 1)
   {
      c = linbuf[k];
      if ((c == '\\') && (linbuf[k + 1] == '\n'))
      {  /* erase escaped newline */
         linbuf[k++] = ' ';
         linbuf[k] = ' ';
         if (inside_comment == 3)
         {
            warn("-- WARNING: comment terminated with escaped newline in `define in '%s' on line %d\n",
                 yyfilename, yylineno+pplineno);
            inside_comment = 0;
         }
      }
      else if ((inside_comment == 0) && (c == '/'))  inside_comment = 1;
      else if ((inside_comment == 1) && (c == '*'))  inside_comment = 2;
      else if ((inside_comment == 1) && (c == '/'))  inside_comment = 3;
      else if ( inside_comment == 1)                 inside_comment = 0;
      else if ((inside_comment == 2) && (c == '*'))  inside_comment = 4;
      else if ((inside_comment == 3) && (c == '\n')) inside_comment = 5;
      else if ((inside_comment == 4) && (c == '/'))  inside_comment = 5;
      else if ((inside_comment == 4) && (c == '*'))  inside_comment = 4;
      else if ( inside_comment == 4)                 inside_comment = 2;
      if (inside_comment >= 2)
         linbuf[k-1] = linbuf[k] = ' ';
      if (inside_comment == 5)
         inside_comment = 0;
   }
   /* At this point comments have been replaced with spaces.
    * Now make *eovptr point to the first of trailing spaces.
    */
   for (k = eov - 1; k >= bov; k -= 1)
      if (linbuf[k] != ' ')
         break;
   *eovptr = k + 1;
}

static int expand_macro(int *jj, int start_of_token, int end_of_token)
{
   int j = *jj, start = *jj, eob, bq;
   struct sym_entry *sym;
   int len = end_of_token - start_of_token;
   const char *s = &linbuf[start_of_token];
   if (cond != COND_ON)
   {
      if (j == 0)
         expbuf[j++] = ' ';     /* make sure at least one char is in expbuf */
      *jj = j;
      return TRUE;
   }
   for (sym = preproc_symtab; sym != 0; sym = sym->next)
      if ((len == sym->symbol_length) &&
          (strncmp(s, sym->symbol, len) == 0))
         break;
   if (sym == 0)
   {
      char name[300+1];
      int l = (len > 300) ? 300 : len;
      strncpy(name,s,l);
      name[l] = 0;
#if 0
      mesg("** Reference to undefined macro `%s in '%s' on line: %d\n",
           name, yyfilename, yylineno+pplineno);
#endif
      eob = j + 1 + len;
      if (eob >= ExpLineSz)
         return FALSE;             /* no space left in expbuf */
      expbuf[eob]=0;
      expbuf[j++]='`';
      strncpy(&expbuf[j], s, len);
      *jj = eob;
      return TRUE;
   }
   eob = j + sym->value_length;
   if (eob >= ExpLineSz)
      return FALSE;             /* no space left in expbuf */
   strcpy(&expbuf[j], sym->value);
   /*
    * Now check for recursive macros:
    */
   while (find_backquote(&bq, start, eob))
   {
      int eos;
      struct sym_entry *sym = find_symbol(&eos, bq + 1, eob);
      if (sym == 0)
         expbuf[bq] = ' ';      /* whipe out backquote */
      else
      {
         if (!replace_symbol(bq, eos, &eob, sym))
            return FALSE;       /* no space left in expbuf */
      }
   }
   *jj = eob;
   return TRUE;
}


static int find_backquote(int *bq, int start, int eob)
{
   int i;
   for (i = start; i < eob; i += 1)
      if (expbuf[i] == '`')
      {
         *bq = i;
         return TRUE;
      }
   return FALSE;
}


static struct sym_entry *find_symbol(int *eos, int start, int eob)
{
   int i = start, sos, len;
   const char *s;
   struct sym_entry *sym;
   while ((i < eob) && is_space(expbuf[i]))
      i += 1;
   if (i >= eob)
   {
      mesg("** Encountered empty macro call during expansion, '%s' line %d\n",
           yyfilename, yylineno+pplineno);
      return 0;
   }
   sos = i;
   while ((i < eob) && is_alpha_num(expbuf[i]))
      i += 1;
   len = i - sos;
   s = &expbuf[sos];
   for (sym = preproc_symtab; sym != 0; sym = sym->next)
      if ((len == sym->symbol_length) &&
          (strncmp(s, sym->symbol, len) == 0))
         break;
   if (sym == 0)
   {
      char name[300+1];
      int l = (len > 300) ? 300 : len;
      strncpy(name,s,l);
      name[l] = 0;
      mesg("** Reference to undefined macro `%s in '%s' on line: %d\n",
           s, yyfilename, yylineno+pplineno);
   }
   *eos = i;
   return sym;
}


static int replace_symbol(int bq, int eos, int *eob, struct sym_entry *sym)
{
   int len1 = eos - bq;
   int len2 = sym->value_length;
   int new_eob = *eob + len2 - len1;
   if (new_eob >= ExpLineSz)
      return FALSE;             /* no space left in expbuf */
   /* first move the buffer content that follows the macro call */
   memmove(/* dest */ &expbuf[eos + len2 - len1],
           /* src  */ &expbuf[eos], *eob - eos);
   /* then fill in the remaining hole */
   strncpy(&expbuf[bq], sym->value, len2);
   *eob = new_eob;
   return TRUE;
}


static int ifdef_macro(int *ii, int *new_cond)
{
   int i = *ii;
   int c, sos, eos, len;
   const char *s;
   struct sym_entry *sym;
   while ((c = linbuf[i]) && is_space(c))
      i += 1;
   if (c == 0)
      return FALSE;             /* try next time */
   sos = i;
   while ((c = linbuf[i]) && is_alpha_num(c))
      i += 1;
   if (c == 0)
      return FALSE;             /* try next time */
   eos = i;
   len = eos - sos;
   if (len == 0)
   {
      mesg("** `ifdef without argument in '%s' on line %d\n",yyfilename, yylineno+pplineno);
      *new_cond = COND_BLOCKED;
      *ii = i;
      return TRUE;
   }
   s = &linbuf[sos];
   for (sym = preproc_symtab; sym != 0; sym = sym->next)
      if ((len == sym->symbol_length) &&
          (strncmp(s, sym->symbol, len) == 0))
         break;
   if (sym == 0)
      *new_cond = COND_OFF;
   else
      *new_cond = COND_ON;
   *ii = i;
   return TRUE;
}
