
/************************************************************
**
**       COPYRIGHT (C) 1988 UNIVERSITY OF PITTSBURGH
**                  ALL RIGHTS RESERVED
**
**        This software is distributed on an as-is basis
**        with no warranty implied or intended.  No author
**        or distributor takes responsibility to anyone 
**        regarding its use of or suitability.
**
**        The software may be distributed and modified 
**        freely for academic and other non-commercial
**        use but may NOT be utilized or included in whole
**        or part within any commercial product.
**
**        This copyright notice must remain on all copies 
**        and modified versions of this software.
**
************************************************************/

/*
**   File:        cmd_misc.c
**
**   Functions:   apply, cmd_alias, cmd_cmdfile, cmd_comment,
**                cmd_excl, cmd_help, cmd_quit, fapply, cmd_what,
**                cmd_timer,
**		  finput
**
**   Description: this file contains miscellaneous commands along with
**                some support routines which are used to execute
**                these (and other) commands.
**
**   Author:      Lyn Ann Mears    (LAM) - University of Pittsburgh, EE Dept.
**   Modified by: Alan R. Martello (ARM) - University of Pittsburgh, EE Dept.
**		  Anthony J. Bybell(AJB) - Penn State University
**
**   Revision History
**   ----------------
**   Date      Initials   Description of change
**
**   thru 7/88   LAM      Initial modification of RSIM along with writing
**                        new routines, including debug of code.
**   08/04/88    ARM      Code cleanup; added comments.
**   08/23/88    ARM      Add alias command.
**   08/24/97    AJB  	  Added aet support.
**
**   Some code taken and adapted from RSIM work performed by Chris Terman.
**
*/


#include "net.h"



/*
**   Function:     apply
**
**   File:         cmd_misc.c
**
**   Called by:    cmd_dopath (cmd_do.c) with 'fun' cpath
**                 cmd_excl (cmd_misc.c) with 'fun' info
**                 cmd_setvalue (cmd_set.c) with 'fun' setin
**                 cmd_quest (cmd_show.c) with 'fun' info
**
**   Description:  apply the given function to each node on the command
**                 line; arguments are checked to verify they reference
**                 some node in the curcuit, patterns are allowed.
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: None.
*/
apply (fun, arg)
    void            (*fun) ();
int             arg;

{
    register int    i;

    for (i = 1; i < targc; i++)
    {
	if (match_net (targv[i], fun, arg) == 0)
	    verror ("cannot find node %s\n", targv[i]);
    }
}



/*
**   Function:     cmd_alias
**
**   File:         cmd_misc.c
**
**   Called by:    invocation of 'alias' command
**
**   Description:  define or display a list of alias'
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
int
cmd_alias ()
{
    int             i, j;

    if (targc == 1)
    {
	/* output known aliases  */
	lprintf (outfile, "alias list:\n");

	for (i = 0; alias_list[i].alias_name[0] != '\0'; i++)
	{
	    lprintf (outfile, "   %s\t%s\n",
		     alias_list[i].alias_name, alias_list[i].actual_cmd);
	}
	if (i == 0)
	{
	    /* no known aliases defined  */
	    lprintf (outfile, "   None.\n");
	}
    }
    else if (targc == 2)
    {
	/* error, we need 3 or more items to define al alias  */
	verror (stderr, "error, invalid alias specification\n");
	verror ("%s\n", USAGE_ALIAS);
    }
    else
    {
	/* define an alias  */

	/* alias is in targv[1], check its size  */
	if (strlen (targv[1]) >= MAX_ALIAS_NAME)
	{
	    verror ("error, alias name '%s' too long\n", targv[1]);
	    verror ("%s\n", USAGE_ALIAS);
	    return (CMD_SAME_LEVEL);
	}

	/* find the index for the alias  */
	for (i = 0; (i < MAX_ALIAS) && (alias_list[i].alias_name[0] != '\0'); i++)
	{
	    if (!strcmp (alias_list[i].alias_name, targv[1]))
	    {
		/* we found a redefinition of an alias name  */
		break;
	    }
	}

	if (i == MAX_ALIAS)
	{
	    /* too many aliases already defined  */
	    fprintf (stderr, "can't define another alias, maximum number of alias = %d\n",
		     MAX_ALIAS);
	    return (CMD_SAME_LEVEL);
	}

	/* all looks good, i = index for alias  */

	/* copy the alias  */
	strcpy (alias_list[i].alias_name, targv[1]);


	/* concatenate the strings together to form the command  */
	strcpy (alias_list[i].actual_cmd, targv[2]);

	for (j = 3; j < targc; j++)
	{
	    strcat (alias_list[i].actual_cmd, " ");
	    strcat (alias_list[i].actual_cmd, targv[j]);
	}
    }
    return (CMD_SAME_LEVEL);
}


/*
**   Function:     cmd_cmdfile
**
**   File:         cmd_misc.c
**
**   Called by:    invocation of '@' command
**
**   Description:  thie routine redirects the input stream
**                 by invoking a command file
**
**   Arguments:    No explicit arguments, an implicit argument, the name
**                 of the command file to invoke passed in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
int
cmd_cmdfile ()
{
    if (targc != 2)
    {
	verror ("wrong number of args to '@' command\n");
	verror ("%s\n", USAGE_CMDFILE);
    }
    else if (!finput (targv[1]))
    {
	verror ("%s%s\n", "cannot open command file specified, ", targv[1]);
	verror ("%s\n", USAGE_CMDFILE);
    }

    return (CMD_SAME_LEVEL);
}


/*
**   Function:     cmd_comment
**
**   File:         cmd_misc.c
**
**   Called by:    invocation of '|' command
**
**   Description:  this routine does nothing but return
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
cmd_comment ()
{
    return (CMD_SAME_LEVEL);
}


/*
**   Function:     cmd_excl
**
**   File:         cmd_misc.c
**
**   Called by:    invocation of '!' command
**
**   Description:  this routine is used to output a list of all
**                 other nodes which effect the specified nodes
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
int
cmd_excl ()
{
    if (targc >= 2)
    {
	apply (info, 1);
    }
    else
    {
	verror ("wrong number of arguments to '!' command\n");
	verror ("%s\n", USAGE_EFFECT_ME);
    }
    return (CMD_SAME_LEVEL);
}


/*
**   Function:     cmd_help
**
**   File:         cmd_misc.c
**
**   Called by:    invocation of 'help' command
**
**   Description:  this routine outputs a list of commands and brief
**                 descriptions of their effect
**
**   Arguments:    None.
**
**   Return Value: CMD_SAME_LEVEL
*/
cmd_help ()
{
    int             i;

    for (i = 0; cmdtbl[i].name != NULL; i++)
    {
	fprintf (outfile, "%-10s  %s\n", cmdtbl[i].name, cmdtbl[i].help_txt);
    }

    return (CMD_SAME_LEVEL);
}


/*
**   Function:     cmd_quit
**
**   File:         cmd_misc.c
**
**   Called by:    invocation of 'quit' command
**
**   Description:  this routine is used to exit the current level of
**                 command processing (enterred by the use of the '@'
**                 command)
**
**   Arguments:    None.
**
**   Return Value: CMD_UP_LEVEL
*/
/* exit this level of command processing */
int
cmd_quit ()
{
    return (CMD_UP_LEVEL);
}


/*
**   Function:     fapply
**
**   File:         cmd_misc.c
**
**   Called by:    cmd_settrace (cmd_set.c), 'fun' of 'xtrace'
**                 cmd_display (cmd_show.c), 'fun' of 'xwatch'
**
**   Description:  apply the given function to each node on the command
**                 line; arguments are checked to verify they reference
**                 some node in the curcuit, patterns are allowed.
**                 pass an argument ('flag') of 1 if the node name was
**                 preceded by a '-' sign, 0 otherwise
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
fapply (fun)
    void            (*fun) ();

{
    register char  *p;
    register int    i, flag;

    for (i = 1; i < targc; i++)
    {
	p = targv[i];
	if (*p == '-')
	{
	    flag = 1;
	    p++;
	}
	else
	    flag = 0;

	if (match_net (p, fun, flag) == 0)
	    verror ("cannot find node %s\n", p);
    }
}


/*
**   Function:     finput
**
**   File:         cmd_misc.c
**
**   Called by:    cmd_cmdfile (cmd_misc.c), main (vsim.c)
**
**   Description:  this routine attempts to open the file name
**                 passed in and resets all global variables to
**                 reference this new stream;  if the file is
**                 openned successfully, 'input_command' is called
**                 to process all commands in the file.
**
**   Arguments:    'name' - the file name to open for commands
**
**   Return Value: 1 if everything is OK, 0 if we have trouble.
*/
int
finput (name)
    char           *name;
{
    FILE           *next;
    int             olineno;
    char           *ofname;

    /* try to modify our current input stream by reading the input file  */
    if (next = fopen (name, "r"))
    {
	/* save the old line number and file name  */
	ofname = filename;
	olineno = lineno;

	/* setup our new line number and counter  */
	filename = name;
	lineno = 0;

	/* process the commands from the new stream  */
	input_command (next);

	/* all finished processing commands, close the stream  */
	fclose (next);

	/* restore the previous file name and line counter  */
	filename = ofname;
	lineno = olineno;

	/* all's well that ends well...  */
	return (1);
    }
    else
    {
	/* trouble in River City... (can't open the file)  */
	return (0);
    }
}

#if INCLUDE_TIMER
/*
**   Function:     cmd_timer
**
**   File:         cmd_misc.c
**
**   Called by:    main 
**
**   Description:  routine to setup/display the timer
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
extern long nevent;         /* number of current event       */
/*  two variables for statistics  */
extern long st_num_events;	/*  starting number of events  */
extern long end_num_events;	/*  ending number of events  */

cmd_timer ()
{


    if (targc == 1)
    {
	/* output current state of timer  */
	lprintf (outfile, "timer is %s\n", 
                 timer_state == TIMER_ON ? TIMER_ON_STR : TIMER_OFF_STR);
    }
    else if (targc == 2)
    {
        /*  check that we have TIMER_ON_STR, TIMER_OFF_STR, or TIMER_SHOW_STR  */
        if (!strcmp (targv[1], TIMER_ON_STR))
        {
            /*  turn the timer on (if not on already)  */
            if (timer_state == TIMER_OFF)
            {
                timer_state = TIMER_ON;
                elapsed_time = 0L;
		st_num_events = nevent;
/*                lprintf (outfile, "timer is reset and running\n"); */
            }
            else
            {
                fprintf (stderr, "timer is already running\n");
            }
        }
        else if (!strcmp (targv[1], TIMER_OFF_STR))
        {
            timer_state = TIMER_OFF;
	    end_num_events = nevent;
/*            lprintf (outfile, "timer is off\n");  */
        }
        else if (!strcmp (targv[1], TIMER_SHOW_STR))
        {
            /*  output the current elapsed time  */

	    /*  in case we ask for the number of events while the timer is running  */
	    if (timer_state == TIMER_ON) end_num_events = nevent;

            lprintf (outfile, "elapsed time is %ld.%03ld sec.\n",
                     (long) (elapsed_time / 1000L) , (long) (elapsed_time % 1000L));
            lprintf (outfile, "number of events = %ld   events/sec = %.2f\n",
		     (long) (end_num_events - st_num_events),
		     (float) ((float) (end_num_events - st_num_events)/ 
			      ((float) elapsed_time) * 1000.0));

            if (timer_state == TIMER_ON)
            {
                lprintf (outfile, "timer is %s\n", TIMER_ON_STR);
            }
        }
        else
        {
            /*  an error  */
            fprintf (stderr, "unrecognized timer keyword, use one of %s, %s, or %s\n",
                     TIMER_ON_STR, TIMER_OFF_STR, TIMER_SHOW_STR);
        }
    }
    else
    {
        /*  an error  */
            fprintf (stderr, "too many arguments to timer, use one of %s, %s, or %s\n",
                     TIMER_ON_STR, TIMER_OFF_STR, TIMER_SHOW_STR);
    }
    return (CMD_SAME_LEVEL);
}    
#endif

/****************************************************************************/

/*
 * aet commands
 */
int aeton=-1, aetoff=-1;
FILE *aethandle=NULL;
char *aetname=NULL;

int cmd_aet(void)
{
if(targc!=3)
	{
        verror ("wrong number of args to 'aet' command\n");
        verror ("%s\n", USAGE_AET);
	}
	else
	{
	if(!strcmp("on",targv[1]))
		{
		if(!aethandle)
			{
			aeton=atoi(targv[2]);			
			if(!(aethandle=fopen(aetname,"wb")))
				{
				verror("%s%s%s\n","could not open aet file, '",
								aetname,"'");
				}
				else
				{
				write_aet_header();
				}
			}
			else
			{
			int t;
			t=atoi(targv[2]);

			verror("%s%d%s%d\n", "aet already set to start at ",
						aeton, " changing to ",t);
			aeton=t;
			}
		}
	else
	if(!strcmp("off",targv[1]))
		{
		aetoff=atoi(targv[2]);			
		}
		else
		{
		verror("%s%s\n", "invalid option in 'aet' command, ",targv[1]);
	        verror ("%s\n", USAGE_AET);
		}
	}
    return (CMD_SAME_LEVEL);
}


void write_aet_header(void)
{
int index;
nptr n;

fprintf(aethandle,"#aet\n");		/* aet symbol table preamble */

for (index = 0; index < HASHSIZE; index++)
	{
        for (n = hash[index]; n != NULL; n = n->hnext)
        	{
		fprintf(aethandle, "%s\n", n->nname);
        	}
    	}
fprintf(aethandle,"#aet\n");		/* end of header */
}


void write_aet_frame(int t)
{
int index;
nptr n;

fprintf(aethandle,"%d\t",t);	/* time from step command */

for (index = 0; index < HASHSIZE; index++)
	{
        for (n = hash[index]; n != NULL; n = n->hnext)
        	{
		fprintf(aethandle, "%c", n->nvalue);
        	}
    	}
fprintf(aethandle, "\n");
}




