
/************************************************************
**
**       COPYRIGHT (C) 1988 UNIVERSITY OF PITTSBURGH
**                  ALL RIGHTS RESERVED
**
**        This software is distributed on an as-is basis
**        with no warranty implied or intended.  No author
**        or distributor takes responsibility to anyone 
**        regarding its use of or suitability.
**
**        The software may be distributed and modified 
**        freely for academic and other non-commercial
**        use but may NOT be utilized or included in whole
**        or part within any commercial product.
**
**        This copyright notice must remain on all copies 
**        and modified versions of this software.
**
************************************************************/

/*
**   File:        cmd_show.c
**
**   Functions:   dvec (static), dnode (static), xwatch (static),
**                cmd_display, cmd_pnlist, cmd_print_node,
**                cmd_quest
**
**   Description:
**		  This file contains 'cmd_XXXXXX' functions which
**                are used to display node informationand
**                all static routines used by those functions.
**                All static routines appear first in the file.
**
**   Author:      Lyn Ann Mears    (LAM) - University of Pittsburgh, EE Dept.
**   Modified by: Alan R. Martello (ARM) - University of Pittsburgh, EE Dept.
**
**   Revision History
**   ----------------
**   Date      Initials   Description of change
**
**   thru 7/88   LAM      Initial modification of RSIM along with writing
**                        new routines, including debug of code.
**   08/04/88    ARM      Code cleanup; added comments.
**   08/23/88    ARM      Add wildcarding function to cmd_print_node
**
**
**   Some code taken and adapted from RSIM work performed by Chris Terman.
**
*/

#include "net.h"

#define TAB_SIZE 8

/*
**   Function:     dvec (static)
**
**   File:         cmd_show.c
**
**   Called by:    cmd_pnlist (cmd_show.c)
**
**   Description:  routine to display the bit vector as a series of
**                 bits and as a decimal integer
**
**   Arguments:    b - pointer to bit vector to display
**
**   Return Value: None.
*/
static void
dvec (b)
    register bptr   b;
{
    long            val = 0;
    register int    vok = 1;
    register int    f, ch;
    int             se; /*  sign extend bit  */

    /*  recall that the bit vector is stored in right to left order  */

    if (strlen(b->bname) < TAB_SIZE)
    {
        lprintf (outfile, "%s\t\t= ", b->bname);
    }
    else
    {
        lprintf (outfile, "%s\t= ", b->bname);
    }

    for (f = b->bnbits-1; f >= 0; f--)
    {
	ch = b->bbits[f]->nvalue;

	lprintf (outfile, "%c", ch);

        if (ch == '1')
        {
            val |= (1 << f);
        }
        else
        {
            if (ch != '0') vok = 0;
        }
    }

    if (vok)
    {
        /*  if we have no undefined bits, output the decimal value  */
	lprintf (outfile, "\t[%lu / 0x%lx]", val, val);

        /*  sign extend the uppermost bit  */
        for (f = b->bnbits-1, se = (val >> f) & 1;
             f < 32; f++)
        {
            val |= (se << f);
        }
	lprintf (outfile, "\t= (%ld)", val);

    }

    lprintf (outfile, "\n");
}


/*
**   Function:     dnode (static)
**
**   File:         cmd_show.c
**
**   Called by:    cmd_pnlist (cmd_show.c)
**
**   Description:  routine to output the current node name and value,
**                 while keeping track of the current column number
**
**   Arguments:    n - pointer to node to output
**
**   Return Value: None.
*/
static void
dnode (n)
    register nptr   n;
{
    /* skip down the alias list  */
    while (n->nflags & ALIAS)
	n = n->nlink;

    /* break the line if necessary  */
    if ((column + strlen (n->nname) + 3) >= MAXCOL)
    {
	lprintf (outfile, "\n");
	column = 0;
    }

    /* output the current node  */
    if (strlen(n->nname) < TAB_SIZE)
    {
        lprintf (outfile, "%s = %c\t\t", n->nname, n->nvalue);
        column += strlen (n->nname) + TAB_SIZE*2;
    }
    else
    {
        lprintf (outfile, "%s = %c\t", n->nname, n->nvalue);
        column += strlen (n->nname) + TAB_SIZE;
    }
}


/*
**   Function:     xwatch (static)
**
**   File:         cmd_show.c
**
**   Called by:    cmd_display (cmd_show.c)
**
**   Description:  routine to delete/insert requested node on display list
**
**   Arguments:    n - pointer to node of interest
**                 flag - 0 => insert node onto display list
**                        not 0 => just delete node from display list
**
**   Return Value: None.
*/
static void
xwatch (n, flag)
    nptr            n;
    int             flag;
{
    /* delete the node from the display list (doesn't hurt if it's not there  */
    idelete (n, &wlist);

    if (!flag)
    {
        /* insert the node onto the display list if requested  */
	iinsert (n, &wlist);
    }
}


/*
**   Function:     cmd_display
**
**   File:         cmd_show.c
**
**   Called by:    invocation of 'watch' command
**
**   Description:  add/remove all requested nodes from display list
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
int
cmd_display ()
{
    iptr            list_ptr;
    register iptr   w, w2;
    register nptr   tmp_nptr;

    if (targc >= 2)
    {
	fapply (xwatch);
    }
    else
    {
	verror ("wrong number of arguments to 'watch' command\n");
	verror ("%s\n", USAGE_WATCH);
    }

    /*  now sort all 'watched' bits  */

    for (w = wlist; (w != NULL) && (w->next != NULL); w = w->next) {
      for(w2 = w->next; w2 != NULL; w2 = w2->next) {
	if (swap_strings(w->inode->nname, w2->inode->nname)) {
	  /*  perform swap  */
	  tmp_nptr = w->inode;
	  w->inode = w2->inode;
	  w2->inode = tmp_nptr;
	}
      }
    }

    return (CMD_SAME_LEVEL);
}


/*
**   Function:     cmd_pnlist
**
**   File:         cmd_show.c
**
**   Called by:    invocation of 'dl' command
**
**   Description:  routine to output all values on the current
**                 display list
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
cmd_pnlist ()
{
    register bptr   b;
    register iptr   w;
    register int    i;

    if (targc == 1)
    {
	/* print values on display list  */

	/* print value of each watched bit vector */
	for (b = blist; b != NULL; b = b->bnext)
	    dvec (b);

	/* print value of each watched node making sure we fit on page */
	column = 0;

	for (w = wlist; w; w = w->next)
	    dnode (w->inode);
    }
    else
    {
	/* print nodes specified on the command line  */

	/* first print any bit vectors the user has specified */
	for (i = 1; i < targc; i++)
        {
	    for (b = blist; b != NULL; b = b->bnext)
            {
		if (strcmp (b->bname, targv[i]) == 0)
                {
		    dvec (b);
                }
            }
        }

	/* then print individual nodes */
	column = 0;
	for (i = 1; i < targc; i++)
        {
	    match_net (targv[i], dnode, 0);
        }
    }

    /*
     * some concluding info indicating current simulated time and the state
     * of the event list.
     */
    if (column)
	lprintf (outfile, "\n");

    lprintf (outfile, "time = %d = %2.1fns", cur_delta,
	     0.1 * (float) cur_delta);

    if (pending)
	lprintf (outfile, "; there are pending events");

    lprintf (outfile, "\n");

    return (CMD_SAME_LEVEL);
}


/*
**   Function:     cmd_print_node
**
**   File:         cmd_show.c
**
**   Called by:    invocation of 'node' command
**
**   Description:  output the specified node, wildcards ok
**                 command format: node [-l] node_name
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
cmd_print_node ()
{
    char           *c_ptr;
    nptr            n;
    int             long_dump, arg_index;

    if ((targc == 3 && strcmp (targv[1], "-l")) || targc > 3 || targc == 1)
    {
	verror ("wrong argument format to 'node' command\n");
	verror ("%s\n", USAGE_NODE);
    }
    else
    {
	/* we have valid parameters, now see load the correct options  */
	if (targc == 3)
	{
	    long_dump = TRUE;
	    arg_index = 2;
	}
	else
	{
	    long_dump = FALSE;
	    arg_index = 1;
	}


	/*
	 * see if there are wildcards in the argument, if so, output all
	 * nodes specified, otherwise output only the specified node
	 */
	for (c_ptr = &targv[arg_index][0]; *c_ptr != '\0'; c_ptr++)
	{
	    if (*c_ptr == '*' || *c_ptr == '?')
	    {
		/* we have a wildcard  */
		break;
	    }
	}

	if (*c_ptr == '\0')
	{
	    /* no wildcards in node string  */
	    if ((n = find (targv[arg_index])) == 0)
	    {
		/* couldn't find explicit node  */
		lprintf (outfile, "can not find node %s\n", targv[arg_index]);
	    }
	    else
	    {
		/* output the single node  */
		prnode (n, long_dump);
	    }
	}
	else
	{
	    /* wildcards in string, output all matching nodes  */
	    match_net (targv[arg_index], prnode, long_dump);
	}

    }
    return (CMD_SAME_LEVEL);
}


/*
**   Function:     cmd_quest
**
**   File:         cmd_show.c
**
**   Called by:    invocation of '?' command
**
**   Description:  routine to print known nodes effected by this node;
**                 wildcards gladly accepted
**
**   Arguments:    No explicit arguments, implicit arguments passed
**                 in 'targv'.
**
**   Return Value: CMD_SAME_LEVEL
*/
int
cmd_quest ()
{
    if (targc >= 2)
    {
	apply (info, 0);
    }
    else
    {
	verror ("wrong number of arguments to '?' command\n");
	verror ("%s\n", USAGE_EFFECT_OTHERS);
    }
    return (CMD_SAME_LEVEL);
}
