
/************************************************************
**
**       COPYRIGHT (C) 1988 UNIVERSITY OF PITTSBURGH
**                  ALL RIGHTS RESERVED
**
**        This software is distributed on an as-is basis
**        with no warranty implied or intended.  No author
**        or distributor takes responsibility to anyone 
**        regarding its use of or suitability.
**
**        The software may be distributed and modified 
**        freely for academic and other non-commercial
**        use but may NOT be utilized or included in whole
**        or part within any commercial product.
**
**        This copyright notice must remain on all copies 
**        and modified versions of this software.
**
************************************************************/

/*
**   File:        nsubrs.c
**
**   Functions:   str_match (static), sym_hash (static),
**                find, init_hash, match_net, numberp, n_insert, pnode,
**                walk_net
**
**   Description:
**
**
**   Author:      Lyn Ann Mears    (LAM) - University of Pittsburgh, EE Dept.
**   Modified by: Alan R. Martello (ARM) - University of Pittsburgh, EE Dept.
**
**   Revision History
**   ----------------
**   Date      Initials   Description of change
**
**   thru 7/88   LAM      Initial modification of RSIM along with writing
**                        new routines, including debug of code.
**   08/04/88    ARM      Code cleanup; added comments.
**
**
**   Some code taken and adapted from RSIM work performed by Chris Terman.
**
*/

#include <ctype.h>
#include "net.h"
#include "system_defs.h"


/*
**   Function:     str_match
**
**   File:         nsubrs.c
**
**   Called by:    match_net (nsubrs.c)
**
**   Description:  routine to see if a pattern (possibly containing '*'
**                 wildcards) matches a given string
**                 will NEVER match SIG_ZERO, SIG_ONE, or STR_DUMMY_SIGNAL
**			(or SIG_X or SIG_Z)
**   Return Value: 1 if the pattern matches the string
**                 0 if the pattern doesn't match the string
*/
static int
str_match (p, s)
    register char  *p, *s;
{
    /*  check if the string is SIG_ZERO, SIG_ONE, or STR_DUMMY_SIGNAL  */
    if (!strcmp (SIG_ZERO, s) || !strcmp (SIG_ONE, s) ||
	!strcmp (SIG_X, s)    || !strcmp (SIG_Z, s)   ||
        !strcmp (STR_DUMMY_SIGNAL, s))
    {
        return (0);
    }

    for (;;)
    {
	if (*p == '*')
	{
	    /* skip past multiple wildcards */
	    while (*p == '*')
		p++;

	    /* if pattern ends with wild card, automatic match */
	    if (*p == '\0')
		return (1);

	    /*
	     * *p now points to first non-wildcard character, find matching
	     * character in string, then recursively match remaining pattern.
	     * if recursive match fails, assume current '*' matches more...
	     */
	    while (*s)
	    {
		while (*s != *p)
		{
		    /*
		     * return if at end of string and we still have pattern
		     * left
		     */
		    if (*s++ == '\0')
			return (0);
		}
		/* attempt to match the new occurrance  */
		if (str_match (p + 1, ++s))
		    return (1);
	    }

	    /*
	     * couldn't find matching character after '*', no match
	     */
	    return (0);
	}
	else if (*p == '\0')
	    return (*s == '\0');
	else if (*p++ != *s++)
	    return (0);
    }
}

/*
**   Function:     sym_hash
**
**   File:         nsubrs.c
**
**   Called by:    n_insert (nsubrs.c), find (nsubrs.c)
**
**   Description:  routine to compute the index of the given 'name' in the
**                 hash table
**
**   Return Value: index into hash table where 'name' should live
*/
static int
sym_hash (name)
    register char  *name;
{
    register int    hashcode = 0;

    for (; *name != '\0'; name++)
    {
	hashcode += *name;
    }

    return (hashcode % HASHSIZE);
}

/*
**   Function:     find
**
**   File:         nsubrs.c
**
**   Called by:    presim, sequ(5), vec, vsim(2)
**
**   Description:  routine to find the node 'name' in the hash table
**
**   Return Value: pointer to node with specified name,
**                 if the named node has al ALIAS list after it,
**                 a pointer to the final (real, non-ALIAS) node.
**                 if the named node is not in the hash table, return NULL.
*/
nptr
find (name)
    register char  *name;
{
    register nptr   ntemp;
    register char  *num = numberp (name);
    int             i;

    for (ntemp = hash[sym_hash (name)]; ntemp != '\0'; ntemp = ntemp->hnext)
    {
	if ((ntemp->nflags & NAMED && !strcmp (name, ntemp->nname)) ||
		num == ntemp->nname)
	{
	    /*
	     * skip ALIAS nodes; these are nodes which are linked after the
	     * current or real node
	     */
	    while (ntemp->nflags & ALIAS)
		ntemp = ntemp->nlink;

	    return (ntemp);
	}
    }
    return (0);
}


/*
**   Function:     init_hash
**
**   File:         nsubrs.c
**
**   Called by:    presim (presim.c)
**
**   Description:  routine to initialize all pointers
**                 in static hash table to NULL
**
**   Return Value: None.
*/
void
init_hash ()
{
    register int    i;

    for (i = 0; i < HASHSIZE; i += 1)
	hash[i] = NULL;
}


/*
**   Function:     match_net
**
**   File:         nsubrs.c
**
**   Called by:    apply (cmd_misc.c), fapply (cmd_misc.c),
**                 cmd_pnlist (cmd_show.c)
**
**   Description:  visit each node in the hash table and call the
**                 function 'fun' with the node as the first argument
**                 and the file pointer, 'f' as the second argument,
**                 only if the pattern 'p' matches the node name
**                 if 'fun' is null, simply count the number of matches
**
**   Return Value: the number of nodes found to match & times 'fun' is called
*/
int
match_net (pattern, fun, arg)
    char           *pattern;
    void            (*fun) ();
int             arg;

{
    register int    index;
    register nptr   n;
    int             total = 0;

    for (index = 0; index < HASHSIZE; index++)
    {
	for (n = hash[index]; n != NULL; n = n->hnext)
	{

	    if (str_match (pattern, pnode (n)))
	    {
		if (fun != NULL)
                {
                    (*fun) (n, arg);
                }
		total++;
	    }
	}
    }

    return (total);
}

/*
**   Function:     numberp
**
**   File:         nsubrs.c
**
**   Called by:    getnode (presim.c)
**
**   Description:  translate an ASCII string into an unsigned integer
**
**   Return Value: value represented by ASCII string, 's'
**                 0 if 's' is not a valid ASCII string
**                 representation of a number
*/
char *
numberp (s)
    register char  *s;
{
    unsigned long  num = 0;

    if (sscanf (s, " %lu", &num) != 1)
	return (0);
    else
	return ((char *)num);
}

/*
**   Function:     n_insert
**
**   File:         nsubrs.c
**
**   Called by:    getnode (presim.c)
**
**   Description:  routine to insert node 'n' in the hash table
**                 at the head of the chain for that entry in
**                 the hash table
**
**   Return Value: None.
*/
void
n_insert (n)
    register nptr   n;
{
    register int    temp;
    register char  *name;

    name = pnode (n);
    temp = sym_hash (name);
    n->hnext = hash[temp];
    hash[temp] = n;
}


/*
**   Function:     pnode
**
**   File:         nsubrs.c
**
**   Called by:    cdoit (cmd_do.c), cpath (cmd_do.c),
**                 cmd_inputs (cmd_show.c), info (network.c)
**                 match_net (nsubrs.c), n_insert (nsubrs.c)
**                 run_vector (vec.c), setin (network.c)
**
**   Description:  return character pointer to ASCII name
**                 representation for node 'n'
**
**   Return Value: pointer to string containing node name  OR
**                 pointer to string containing node number (in ASCII)  OR
**                 pointer to string containing "(null)" if 'n' is NULL
*/
char           *
pnode (n)
    register nptr   n;
{
    static char     number[50];

    if (n == NULL)
	return ("(null)");

    if (n->nflags & NAMED)
	return (n->nname);

    sprintf (number, "%ld", (long) n->nname);

    return (number);
}


