
/************************************************************
**
**       COPYRIGHT (C) 1988 UNIVERSITY OF PITTSBURGH
**                  ALL RIGHTS RESERVED
**
**        This software is distributed on an as-is basis
**        with no warranty implied or intended.  No author
**        or distributor takes responsibility to anyone 
**        regarding its use of or suitability.
**
**        The software may be distributed and modified 
**        freely for academic and other non-commercial
**        use but may NOT be utilized or included in whole
**        or part within any commercial product.
**
**        This copyright notice must remain on all copies 
**        and modified versions of this software.
**
************************************************************/

/*
**   File:        vsim/proc_helper.c
**
**   Functions:   sig_2_int, int_2_sig, sig_assign,
**                cg_not, cg_null, cg_and, cg_or, cg_nand, cg_nor,
**                cg_xor, cg_concat, cg_rising, cg_falling
**                get_next_buffer (static)
**
**   Description: 'helper' routines for process statements
**                
**
**   Author:      Alan R. Martello (ARM) - University of Pittsburgh, EE Dept.
**
**   Revision History
**   ----------------
**   Date      Initials   Description of change
**
**   10/30/88    ARM      Initial writing and debugging.
**    6/23/89    ARM      Add support for 'rising and 'falling gates
**
**
*/

#include <stdio.h>
#include "system_defs.h"
#include "proc_helper.h"
#include "net.h"

/*
**   Function:     get_next_buffer
**
**   File:         vsim/proc_helper.c
**
**   Description:  routine to return the next usable buffer
**
**   Arguments:    None.
**
**   Return Value: pointer to free buffer
*/
static char *
get_next_buffer()
{
#define MAX_BUFFERS  100
#define BUFFER_SIZE  MAX_BIT_NUMBER

    static int i;
    static char temp_buffers[MAX_BUFFERS][MAX_BIT_NUMBER];

    if (++i >= MAX_BUFFERS) 
    {
        i = 0;
    }
    
    return (&(temp_buffers[i][0]));
}    


/*
**   Function:     sig_2_int
**
**   File:         vsim/proc_helper.c
**
**   Description:  routine to extract a portion of a 
**                 signal, convert it to an integer,
**                 and return it as its value
**
**   Arguments:    in1 - pointer to signal to convert
**                 up1 - upper limit of signal
**                 lo1 - lower limit of signal
**
**   Return Value: integer value of signal portion
*/
int sig_2_int (in1, up1, lo1)
    char *in1;
    int up1, lo1;
{
    int val = 0;
    int i;
    
    for (i = up1; i >= lo1; i--)
    {
        if (in1[i] == '0' || in1[i] == '1')
        {
            val = (val << 1) | (in1[i] - '0');
        }
        else
        {
            /*  assume a '0' if not '1' or '0'  */
            val <<= 1;
        }
    }
    
    return (val);
}


/*
**   Function:     int_2_sig
**
**   File:         vsim/proc_helper.c
**
**   Description:  routine to convert an integer passed
**                 in into a binary character array
**
**   Arguments:    val - integer value to convert
**
**   Return Value: pointer to static character array
**                 containing the result of the conversion
*/
char *
int_2_sig (val)
    unsigned int val;
{
    char *ch_val = get_next_buffer();
    int i;

    /*  translate the array  */
    for (i = 0; val != 0; i++, val >>= 1)
    {
        if (val & 1)
        {
            ch_val[i] = '1';
        }
        else
        {
            ch_val[i] = '0';
        }
    }

    /*  zero out the remainder of the places  */
    for (; i < MAX_BIT_NUMBER; i++)
    {
        ch_val[i] = '0';
    }        

    return (&(ch_val[0]));
}



/*
**   Function:     sig_assign
**
**   File:         vsim/proc_helper.c
**
**   Description:  routine to assign one signal range to
**                 another signal range
**
**   Arguments:    in1 - pointer to destination signal
**                 up1 - upper limit of destination signal
**                 lo1 - lower limit of destination signal
**                 in2 - pointer to source signal
**                 up2 - upper limit of source signal
**                 lo2 - lower limit of source signal
**
**   Return Value: None.
*/
void
sig_assign (in1, up1, lo1, in2, up2, lo2)
    char *in1, *in2;
    int up1, lo1, up2, lo2;
{
    if ((up1-lo1) != (up2-lo2))
    {
        fprintf (stderr, "process: (warning) size mismatch on attempt to assign signal\n");
        fprintf (stderr, "process: (warning) forcing lhs to be of size %d-%d\n",
                 up1, lo1);
        up2 = up1;
        lo2 = lo1;
    }
    
    for (;up1 >= lo1 && up2 >= lo2; lo1++, lo2++)
    {
        /*  perform the copy  */
        in1[lo1] = in2[lo2];
    }
}


/*
**   Function:     cg_not
**
**   File:         vsim/proc_helper.c
**
**   Description:  routine to 'not' the signal passed in 
**
**   Arguments:    in1 - pointer to signal to 'not'
**                 up1 - upper limit of signal to 'not'
**                 lo1 - lower limit of signal to 'not'
**
**   Return Value: pointer to static character array
**                 containing the result of the operation
*/
char *
cg_not (in1, up1, lo1)
    char *in1;
    int up1, lo1;
{
    char *ch_val = get_next_buffer();
    register int i,j;

    for (i = 0, j = lo1; j <= up1; j++, i++)
    {
        if (in1[j] == '1')
        {
            ch_val[i] = '0';
        }
        else
        {
            /*  default case  */
            ch_val[i] = '1';
        }
    }        
    
    /*  zero out the rest of the array  */
    for (; i < MAX_BIT_NUMBER; i++)
    {
        ch_val[i] = '0';
    }

    return (ch_val);
}


/*
**   Function:     cg_null
**
**   File:         vsim/proc_helper.c
**
**   Description:  routine to pass the input signal
**
**   Arguments:    in1 - pointer to signal to 'null'
**                 up1 - upper limit of signal to 'null'
**                 lo1 - lower limit of signal to 'null'
**
**   Return Value: pointer to static character array
**                 containing the result of the operation
*/
char *
cg_null (in1, up1, lo1)
    char *in1;
    int up1, lo1;
{
    char *ch_val = get_next_buffer();
    register int i,j;

    for (i = 0, j = lo1; j <= up1; j++, i++)
    {
        if (in1[j] == '1')
        {
            ch_val[i] = '1';
        }
        else
        {
            /*  default case  */
            ch_val[i] = '0';
        }
    }        
    
    /*  zero out the rest of the array  */
    for (; i < MAX_BIT_NUMBER; i++)
    {
        ch_val[i] = '0';
    }

    return (ch_val);
}



/*
**   Function:     cg_and
**
**   File:         vsim/proc_helper.c
**
**   Description:  routine to perform an 'and'
**                 of the two input signals
**
**   Arguments:    in1 - pointer to first operand
**                 up1 - upper limit of first operand
**                 lo1 - lower limit of first operand
**                 in2 - pointer to second operand
**                 up2 - upper limit of second operand
**                 lo2 - lower limit of second operand
**
**   Return Value: pointer to static character array
**                 containing the result of the operation
*/
char *
cg_and (in1, up1, lo1, in2, up2, lo2)
    char *in1, *in2;
    int up1, lo1, up2, lo2;
{
    char *ch_val = get_next_buffer();
    register int i,j,k;

    if ((up1-lo1) != (up2-lo2))
    {
        fprintf (stderr, "process: (warning) size mismatch on 'and' operation\n");
    }

    for (i = 0, j = lo1, k = lo2; 
         j <= up1; j++, i++, k++)
    {
        if (in1[j] == '1' && in2[k] == '1')
        {
            ch_val[i] = '1';
        }
        else
        {
            /*  default case  */
            ch_val[i] = '0';
        }
    }        
    
    /*  zero out the rest of the array  */
    for (; i < MAX_BIT_NUMBER; i++)
    {
        ch_val[i] = '0';
    }

    return (ch_val);
}


/*
**   Function:     cg_or
**
**   File:         vsim/proc_helper.c
**
**   Description:  routine to perform an 'or'
**                 operation on the two input signals
**
**   Arguments:    in1 - pointer to first operand
**                 up1 - upper limit of first operand
**                 lo1 - lower limit of first operand
**                 in2 - pointer to second operand
**                 up2 - upper limit of second operand
**                 lo2 - lower limit of second operand
**
**   Return Value: pointer to static character array
**                 containing the result of the operation
*/
char *
cg_or (in1, up1, lo1, in2, up2, lo2)
    char *in1, *in2;
    int up1, lo1, up2, lo2;
{
    char *ch_val = get_next_buffer();
    register int i,j,k;

    if ((up1-lo1) != (up2-lo2))
    {
        fprintf (stderr, "process: (warning) size mismatch on 'or' operation\n");
    }

    for (i = 0, j = lo1, k = lo2; 
         j <= up1; j++, i++, k++)
    {
        if (in1[j] == '1' || in2[k] == '1')
        {
            ch_val[i] = '1';
        }
        else
        {
            /*  default case  */
            ch_val[i] = '0';
        }
    }        
    
    /*  zero out the rest of the array  */
    for (; i < MAX_BIT_NUMBER; i++)
    {
        ch_val[i] = '0';
    }

    return (ch_val);
}

/*
**   Function:     cg_nand
**
**   File:         vsim/proc_helper.c
**
**   Description:  routine to perform a 'nand' operation
**                 on the two input signals
**
**   Arguments:    in1 - pointer to first operand
**                 up1 - upper limit of first operand
**                 lo1 - lower limit of first operand
**                 in2 - pointer to second operand
**                 up2 - upper limit of second operand
**                 lo2 - lower limit of second operand
**
**   Return Value: pointer to static character array
**                 containing the result of the operation
*/
char *
cg_nand (in1, up1, lo1, in2, up2, lo2)
    char *in1, *in2;
    int up1, lo1, up2, lo2;
{
    char *ch_val = get_next_buffer();
    register int i,j,k;

    if ((up1-lo1) != (up2-lo2))
    {
        fprintf (stderr, "process: (warning) size mismatch on 'nand' operation\n");
    }

    for (i = 0, j = lo1, k = lo2; 
         j <= up1; j++, i++, k++)
    {
        if (in1[j] == '1' && in2[k] == '1')
        {
            ch_val[i] = '0';
        }
        else
        {
            /*  default case  */
            ch_val[i] = '1';
        }
    }        
    
    /*  zero out the rest of the array  */
    for (; i < MAX_BIT_NUMBER; i++)
    {
        ch_val[i] = '0';
    }

    return (ch_val);
}

/*
**   Function:     cg_nor
**
**   File:         vsim/proc_helper.c
**
**   Description:  routine to perform a 'nor' funtion
**                 on the two inputs
**
**   Arguments:    in1 - pointer to first operand
**                 up1 - upper limit of first operand
**                 lo1 - lower limit of first operand
**                 in2 - pointer to second operand
**                 up2 - upper limit of second operand
**                 lo2 - lower limit of second operand
**
**   Return Value: pointer to static character array
**                 containing the result of the operation
*/
char *
cg_nor (in1, up1, lo1, in2, up2, lo2)
    char *in1, *in2;
    int up1, lo1, up2, lo2;
{
    char *ch_val = get_next_buffer();
    register int i,j,k;

    if ((up1-lo1) != (up2-lo2))
    {
        fprintf (stderr, "process: (warning) size mismatch on 'nor' operation\n");
    }

    for (i = 0, j = lo1, k = lo2; 
         j <= up1; j++, i++, k++)
    {
        if (in1[j] == '1' || in2[k] == '1')
        {
            ch_val[i] = '0';
        }
        else
        {
            /*  default case  */
            ch_val[i] = '1';
        }
    }        
    
    /*  zero out the rest of the array  */
    for (; i < MAX_BIT_NUMBER; i++)
    {
        ch_val[i] = '0';
    }

    return (ch_val);
}

/*
**   Function:     cg_xor
**
**   File:         vsim/proc_helper.c
**
**   Description:  routine to perform an 'xor' function
**                 on the two input operands
**
**   Arguments:    in1 - pointer to first operand
**                 up1 - upper limit of first operand
**                 lo1 - lower limit of first operand
**                 in2 - pointer to second operand
**                 up2 - upper limit of second operand
**                 lo2 - lower limit of second operand
**
**   Return Value: pointer to static character array
**                 containing the result of the operation
*/
char *
cg_xor (in1, up1, lo1, in2, up2, lo2)
    char *in1, *in2;
    int up1, lo1, up2, lo2;
{
    char *ch_val = get_next_buffer();
    register int i,j,k;

    if ((up1-lo1) != (up2-lo2))
    {
        fprintf (stderr, "process: (warning) size mismatch on 'xor' operation\n");
    }

    for (i = 0, j = lo1, k = lo2; 
         j <= up1; j++, i++, k++)
    {
        if ((in1[j] == '1' && in2[k] == '0') ||
            (in1[j] == '0' && in2[k] == '1'))
        {
            ch_val[i] = '1';
        }
        else
        {
            /*  default case  */
            ch_val[i] = '0';
        }
    }        
    
    /*  zero out the rest of the array  */
    for (; i < MAX_BIT_NUMBER; i++)
    {
        ch_val[i] = '0';
    }

    return (ch_val);
}

/*
**   Function:     cg_concat
**
**   File:         vsim/proc_helper.c
**
**   Description:  routine to perform a concatenation of
**                 the two input signals
**
**   Arguments:    in1 - pointer to first operand
**                 up1 - upper limit of first operand
**                 lo1 - lower limit of first operand
**                 in2 - pointer to second operand
**                 up2 - upper limit of second operand
**                 lo2 - lower limit of second operand
**
**   Return Value: pointer to static character array
**                 containing the result of the operation
*/
char *
cg_concat (in1, up1, lo1, in2, up2, lo2)
    char *in1, *in2;
    int up1, lo1, up2, lo2;
{
    char *ch_val = get_next_buffer();
    register int i,j;

    /*  start with the second operand  */
    for (i = 0, j = lo2;
         j <= up2; j++, i++)
    {
        ch_val[i] = in2[j];
    }

    /*  then the first operand  */
    for (j = lo1;
         j <= up1; j++, i++)
    {
        ch_val[i] = in1[j];
    }
    
    /*  zero out the rest of the array  */
    for (; i < MAX_BIT_NUMBER; i++)
    {
        ch_val[i] = '0';
    }

    return (ch_val);
}



/*
**   Function:     cg_rising
**
**   File:         vsim/proc_helper.c
**
**   Description:  routine test for a rising condition
**
**   Arguments:    in1 - pointer to string to test
**                 up1 - bit to test (should be zero)
**                 lo1 - bit to test (should be zero)
**
**   Return Value: pointer to static character array
**                 containing the result of the operation
*/
char *
cg_rising (in1, up1, lo1)
    char **in1;
    int up1, lo1;
{
    char *ch_val = get_next_buffer();
    nptr  np;
    register int i;

    if (up1 != 0 || lo1 != 0)
    {
        fprintf (stderr, "internal error, 'rising should only be on a bit\n");
        exit (1);
    }

    /* HACK to get a pointer to the structure  */
    np = (nptr) (* ((long *)((long) in1 + sizeof(char *))));

    /*
    **  test what the current state is and the time of the last
    **  transaction to determine if we are rising or falling
    */

    if ((np->delta_stamp == delta_timestamp) && 
	(np->nflags & FLAG_GATE_RISING)) {

        /*  we are 'rising now  */
        ch_val[0] = '1';
    }
    else
    {
        ch_val[0] = '0';
    }
    
    /*  zero out the rest of the array  */
    for (i = 1; i < MAX_BIT_NUMBER; i++)
    {
        ch_val[i] = '0';
    }

    return (ch_val);
}


/*
**   Function:     cg_falling
**
**   File:         vsim/proc_helper.c
**
**   Description:  routine test for a falling condition
**
**   Arguments:    in1 - pointer to string to test
**                 up1 - bit to test (should be zero)
**                 lo1 - bit to test (should be zero)
**
**   Return Value: pointer to static character array
**                 containing the result of the operation
*/
char *
cg_falling (in1, up1, lo1)
    char **in1;
    int up1, lo1;
{
    char *ch_val = get_next_buffer();
    nptr  np;
    register int i;

    if (up1 != 0 || lo1 != 0)
    {
        fprintf (stderr, "internal error, 'falling should only be on a bit\n");
        exit (1);
    }

    /* HACK to get a pointer to the structure  */
    np = (nptr) (* ((long *)((long) in1 + sizeof(char *))));

    /*
    **  test what the current state is and the time of the last
    **  transaction to determine if we are rising or falling
    */
    if ((np->delta_stamp == delta_timestamp) && 
	(np->nflags & FLAG_GATE_FALLING)) {

        /*  we are 'falling now  */
        ch_val[0] = '1';
    }
    else
    {
        ch_val[0] = '0';
    }
    
    /*  zero out the rest of the array  */
    for (i = 1; i < MAX_BIT_NUMBER; i++)
    {
        ch_val[i] = '0';
    }

    return (ch_val);
}


