
/************************************************************
**
**       COPYRIGHT (C) 1988 UNIVERSITY OF PITTSBURGH
**                  ALL RIGHTS RESERVED
**
**        This software is distributed on an as-is basis
**        with no warranty implied or intended.  No author
**        or distributor takes responsibility to anyone 
**        regarding its use of or suitability.
**
**        The software may be distributed and modified 
**        freely for academic and other non-commercial
**        use but may NOT be utilized or included in whole
**        or part within any commercial product.
**
**        This copyright notice must remain on all copies 
**        and modified versions of this software.
**
************************************************************/

/*
**   File:        vec.c
**
**   Functions:   define_vector (static), get_vector_value (static),
**                free_vector, free_vlist,
**                cycle, clear_vectors, run_vector, set_vector
**
**
**   Description:
**
**
**   Author:      Lyn Ann Mears    (LAM) - University of Pittsburgh, EE Dept.
**   Modified by: Alan R. Martello (ARM) - University of Pittsburgh, EE Dept.
**
**   Revision History
**   ----------------
**   Date      Initials   Description of change
**
**   thru 7/88   LAM      Initial modification of RSIM along with writing
**                        new routines, including debug of code.
**   08/04/88    ARM      Code cleanup; added comments.
**   08/25/88    ARM      Repartition vector routines to utilize
**                        common functions for both clock and user
**                        vector handling
**
**
**   Some code taken and adapted from RSIM work performed by Chris Terman.
**
*/

#include "net.h"


 /* vector functions for vsim */
 /*
  * Vectors are defined as a sequence of values for a given node.  A list of
  * vectors may be strung together to form a vector list.  This vector list
  * is used to set a group of inputs to certain values at the same time.  For
  * example, if there are three vectors, V1[] attached to NODE_1, V2[]
  * attached to NODE_2, and V3[] attached to NODE_3, then at some time, T,
  * NODE_1 has the value of V1[T], NODE_2 has the value of V2[T], and NODE_3
  * has the value of V3[T].
  * 
  * Vectors may be made up of the following values:  '1' - a logical high, '0' -
  * a logical low, 'x' or 'z' - logically undefined,
  * and '*' meaning to repeat the
  * preceding value indefinitely. The '*' must be the last character in a
  * vector string.
  * 
  * The length of the vectors on a vector list do not necessarily have to be the
  * same.  If the simulator attempts to use an element (for example the I-th
  * element) in a vector with only J elements, the value used is defined as
  * either the final value if the vector ended in a '*', or as the (I mod
  * J)-th element, for a vector with a finite number of elements which does
  * not contain a '*' as its final character.
  * 
  * There are two vector lists supported at this time.  One vector list is for
  * implementing a clocking function (via the 'clock' command), and the
  * second vector function is for implementing a list of user-defined
  * vectors.
  */

 /*
  * length of the longest clock vector, maximum length of a user-defined
  * vector
  */
static int      max_len_user = 0;
static int      max_len_clock = 0;

/*
**   Function:     define_vector (static)
**
**   File:         vec.c
**
**   Called by:    define_clock (vec.c), set_vector (vec.c)
**
**   Description:  routine to allocate memory for the vector and
**                 parse the vector string, initilizing the
**                 vector node as requested
**
**   Arguments:    node_name - the name of the node this vector
**                             is defined for
**                 vector_string - the string of values for the vector
**
**   Return Value: address of vector structure created,
**                 NULL if error creating structure
**                 (i.e. can't find the node name specified)
*/
static Vector  *
define_vector (node_name, vector_string)
    char           *node_name;
    char           *vector_string;
{
    nptr            node_ptr;
    char            ch;
    Vector         *vector;

    /* attempt to find the user-specified node name to tie this vector to  */
    node_ptr = find (node_name);

    /*
     * if the user specified a non-existant node name, inform them of the
     * error in their ways
     */
    if (node_ptr == NULL)
    {
	verror ("can't find vector node %s\n", node_name);
	return (NULL);
    }

    /*
     * allocate space for the vector node, its name, and its array of values
     * (the actual vector itself)
     */

    vector = (Vector *) malloc (sizeof (Vector));
    vector->vec_values = (char *) malloc (sizeof (char) * (strlen (vector_string) + 1));

    /* copy the name in and initialize the other fields  */
    vector->node = node_ptr;
    vector->repeat_count = 0;
    vector->num_elements = 0;

    while (ch = *vector_string++)
    {
	switch (ch)
	{
	    /* a high value in the vector  */
	  case '1':
	    vector->vec_values[vector->num_elements++] = '1';
	    break;

	    /* a low value in the vector  */
	  case '0':
	    vector->vec_values[vector->num_elements++] = '0';
	    break;

	    /* a don't care value in the vector  */
	  case 'x':
	    vector->vec_values[vector->num_elements++] = 'x';
	    break;

	    /* a high impedance value in the vector  */
	  case 'z':
	    vector->vec_values[vector->num_elements++] = 'z';
	    break;

	  case '*':		/* '*' means repeat last setting indefinitely */
	    if (vector->num_elements == 0)
	    {
		verror ("bad vector, contains only %c\n", ch);
		free_vector (vector);
		return (NULL);
	    }

	    vector->vec_values[vector->num_elements++] = '*';
	    vector->repeat_count = 1;
	    break;

	  default:
	    verror ("unknown state type (%c) in vector string\n",ch);
	    verror ("vector string must contain '1', '0', 'x', 'z', or end with '*'\n");
	    free_vector (vector);
	    return (NULL);
	}
    }

    /* nul-terminate the vector string  */
    vector->vec_values[vector->num_elements] = '\0';

    return (vector);
}


/*
**   Function:     get_vector_value (static)
**
**   File:         vec.c
**
**   Called by:    cycle (vec.c), run_vector (vec.c)
**
**   Description:  routine to return the n-th value
**                 for the vector passed in; check for
**                 wildcards (which implies repeat the last
**                 element of the vector); if the element
**                 requested is not within the vector,
**                 'fold' the request by performing a
**                 modulus of the vector length
**
**   Arguments:    vector - a pointer to the current Vector
**                          structure
**                 req_element - an integer indicating the
**                               n-th vector value to return
**
**   Return Value: the vector value
*/
static char
get_vector_value (vector, req_element)
    Vector         *vector;
    int             req_element;
{
    char            ch;

    /*
     * if we have not reached the end of the vector, return the requested
     * value
     */
    if (req_element < vector->num_elements)
    {
	ch = vector->vec_values[req_element];
    }
    else
    {
	/* our requested value exceeds the vector length, decide what to do  */
	if (vector->repeat_count == TRUE)
	{
	    /*
	     * if we have a repeated value, simply return the last value in
	     * the vector string
	     */
	    ch = vector->vec_values[vector->num_elements - 1];
	}
	else
	{
	    /*
	     * if we don't have a repeated value, simply perform a modulus on
	     * the requested value to return a value which lies within the
	     * vector
	     */
	    ch = vector->vec_values[req_element % (vector->num_elements)];
	}
    }
    return (ch);
}




/*
**   Function:     add_to_vector_list
**
**   File:         vec.c
**
**   Called by:
**
**   Description:  routine to add a specified vector to one
**                 of the two vector lists
**
**   Arguments:    vec_list - either VLIST_CLOCK or VLIST_USER
**                            indicating which vector list the
**                            vector should be added to
**                 node_name - the new node name to be added
**                 vec_value - the value of the vector
**
**   Return Value: None.
*/
void
add_to_vector_list (vec_list, node_name, vec_value)
    int             vec_list;
    char           *node_name;
    char           *vec_value;
{
    register int    i;
    register Vector **list_ptr, *v;
    register int   *max_len_ptr;

    switch (vec_list)
    {
      case VLIST_CLOCK:
	list_ptr = &vec_clock;
	max_len_ptr = &max_len_clock;
	break;

      case VLIST_USER:
	list_ptr = &vec_user;
	max_len_ptr = &max_len_user;
	break;
    }

    if ((v = define_vector (node_name, vec_value)) != NULL)
    {
	v->next = *list_ptr;
	*list_ptr = v;
	*max_len_ptr = MAX (*max_len_ptr, strlen (vec_value));
    }
}


/*
**   Function:     cycle
**
**   File:         vec.c
**
**   Called by:
**
**   Description:  routine to cycle the entire network
**                 through a single clock cycle (defined
**                 by the number of vectors which define
**		   the clock and the length of each vector)
**
**   Arguments:    None.
**
**   Return Value: None.
*/
cycle ()
{
    register int    i;
    register Vector *v;

    /* run the clock for the maximum number of cycles  */
    for (i = 0; i < max_len_clock; i += 1)
    {
	for (v = vec_clock; v != NULL; v = v->next)
	    setin (v->node, get_vector_value (v, i));
	step (cur_delta + stepsize);
    }
}


/*
**   Function:     free_vlist
**
**   File:         vec.c
**
**   Called by:    clear_vectors (vec.c), define_clock (vec.c)
**
**   Description:  routine to free a linked list of vectors
**
**   Arguments:    vec_list - a pointer to a pointer to the
**                            head of the vector list to free
**
**   Return Value: None.
*/
void
free_vlist (vec_list)
    Vector        **vec_list;
{
    register Vector *v, *temp;

    /*  clear the correct length  */
    if (vec_list == &vec_clock)
    {
        max_len_clock = 0;
    }
    else if (vec_list == &vec_user)
    {
        max_len_user = 0;
    }
    else
    {
        fprintf (stderr, "Internal error, cannot find vector length to clear\n");
    }
    
    for (v = *vec_list; v != NULL; v = temp)
    {
	temp = v->next;
	free_vector (v);
    }

    *vec_list = NULL;
}


/*
**   Function:     free_vector
**
**   File:         vec.c
**
**   Called by:
**
**   Description:  routine to free a single vector
**
**   Arguments:    v - the vector to free
**
**   Return Value: None.
*/
free_vector (v)
    Vector         *v;
{
    /* free the value array  */
    free (v->vec_values);

    /* free the vector structure itself  */
    free ((char *) v);
}

/*
**   Function:     run_vector
**
**   File:         vec.c
**
**   Called by:    cmd_vec_run (cmd_vec.c)
**
**   Description:  this routine runs the simulator and
**                 collects all the data specified in the
**                 user vectors, and outputs these results
**
**   Arguments:    num_cycle_str - the number of clock cycles to
**                                 run the simulator
**
**   Return Value: None.
*/
void
run_vector (num_cycle_str)
    char           *num_cycle_str;
{
    register int    num, i, length;
    register iptr   w;
    register Vector *v;
    register char  *p;
    char           *data;

    /* figure out how many outputs we are going to collect */
    if (num_cycle_str == NULL)
    {
	if ((length = max_len_user) == 0)
	{
	    verror ("cannot 'runvec', no vectors defined!\n");
	    return;
	}
    }
    else
    {
	length = atoi (num_cycle_str);
	if (length < 1)
	{
	    verror ("Bad argument to runvec: %s\n", num_cycle_str);
	    return;
	}
    }

    /* count how many outputs there are each cycle, allocate array */
    for (num = 0, w = wlist; w != NULL; w = w->next, num++);

    /* allocate the data buffer to use to store the cycle  */
    data = (char *) malloc (num * (length + 1));

    /* do simulation, collecting outputs as we go */
    for (i = 0; i < length; i += 1)
    {
	/* set the input nodes to the i-th vector value  */
	for (v = vec_user; v != NULL; v = v->next)
	    setin (v->node, get_vector_value (v, i));

	/* perform a single step  */
	step (cur_delta + stepsize);

	/* save all the watched nodes  */
	for (p = &data[i], w = wlist;
		w != NULL; w = w->next, p += length + 1)
	{
	    *p = w->inode->nvalue;
	}

	/* cycle the clock  */
	if (i < length - 1)
	    cycle ();
    }

    /* show data to user */
    for (p = data, w = wlist; w != NULL; w = w->next, p += length + 1)
    {
	p[length] = '\0';
	lprintf (outfile, "%s\t= %s\n", pnode (w->inode), p);
    }
    
    /* turn off the input bit for all the input nodes, previously set by 'setin' */
    for (v = vec_user; v != NULL; v = v->next)
    {
        v->node->nflags &= ~FORCE_VALUE;
    }

    /* free the dynamically allocated data buffer  */
    free (data);
}

void
dump_vlist (vec_list)
    int             vec_list;
{
    char           *header_string;
    Vector         *list_ptr;

    switch (vec_list)
    {
      case VLIST_CLOCK:	/* display the clock vector list  */
	header_string = "clock vector list:\n";
	list_ptr = vec_clock;
	break;

      case VLIST_USER:		/* display the user vector list  */
	header_string = "user vector list:\n";
	list_ptr = vec_user;
	break;
    }

    lprintf (outfile, header_string);

    if (list_ptr == NULL)
    {
	lprintf (outfile, "   None.\n");
    }
    else
    {
	for (; list_ptr != NULL; list_ptr = list_ptr->next)
	{
	    lprintf (outfile, "   %s  -  %s\n",
		     list_ptr->node->nname, list_ptr->vec_values);
	}
    }
}
